"""
====================================
Node API (:class:`reservoirpy.Node`)
====================================

Note
----

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`
- **Learn how to subclass Node to make your own**: :ref:`create_new_node`


**Simple tools for complex reservoir computing architectures.**

The Node API features a simple implementation of computational graphs, similar
to what can be found in other popular deep learning and differentiable calculus
libraries. It is however simplified and made the most flexible possible by
discarding the useless "fully differentiable operations" functionalities. If
you wish to use learning rules making use of chain rule and full
differentiability of all operators, ReservoirPy may not be the tool you need
(actually, the whole paradigm of reservoir computing might arguably not be the
tool you need).

The Node API is composed of a base :py:class:`Node` class that can be described
as a stateful recurrent operator able to manipulate streams of data. A
:py:class:`Node` applies a `forward` function on some data, and then stores the
result in its `state` attribute. The `forward` operation can be a function
depending on the data, on the current `state` vector of the Node, and
optionally on data coming from other distant nodes `states` though feedback
connections (distant nodes can be reached using the `feedback` attribute of the
node they are connected to).

Nodes can also be connected together to form a :py:class:`Model`. Models hold
references to the connected nodes and make data flow from one node to
the next, allowing to create *deep* models and other more complex
architectures and computational graphs.
:py:class:`Model` is essentially a subclass of :py:class:`Node`,
that can also be connected to other nodes and models.

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`
- **Learn how to subclass Node to make your own**: :ref:`create_new_node`

.. currentmodule:: reservoirpy.node

.. autoclass:: Node

   .. rubric:: Methods

   .. autosummary::

      ~Node.call
      ~Node.clean_buffers
      ~Node.copy
      ~Node.create_buffer
      ~Node.feedback
      ~Node.fit
      ~Node.get_buffer
      ~Node.get_param
      ~Node.initialize
      ~Node.initialize_buffers
      ~Node.initialize_feedback
      ~Node.link_feedback
      ~Node.partial_fit
      ~Node.reset
      ~Node.run
      ~Node.set_buffer
      ~Node.set_feedback_dim
      ~Node.set_input_dim
      ~Node.set_output_dim
      ~Node.set_param
      ~Node.set_state_proxy
      ~Node.state
      ~Node.state_proxy
      ~Node.train
      ~Node.with_feedback
      ~Node.with_state
      ~Node.zero_feedback
      ~Node.zero_state

   .. rubric:: Attributes

   .. autosummary::

      ~Node.feedback_dim
      ~Node.fitted
      ~Node.has_feedback
      ~Node.hypers
      ~Node.input_dim
      ~Node.is_fb_initialized
      ~Node.is_initialized
      ~Node.is_trainable
      ~Node.is_trained_offline
      ~Node.is_trained_online
      ~Node.name
      ~Node.output_dim
      ~Node.params

References
==========

    ReservoirPy Node API was heavily inspired by Explosion.ai *Thinc*
    functional deep learning library [1]_, and *Nengo* core API [2]_.
    It also follows some *scikit-learn* schemes and guidelines [3]_.

    .. [1] `Thinc <https://thinc.ai/>`_ website
    .. [2] `Nengo <https://www.nengo.ai/>`_ website
    .. [3] `scikit-learn <https://scikit-learn.org/stable/>`_ website

"""

# Author: Nathan Trouvain at 22/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from contextlib import contextmanager
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Union

import numpy as np
from scipy.sparse import issparse

from ._base import DistantFeedback, _Node, call, check_one_sequence, check_xy, train
from .model import Model
from .type import (
    BackwardFn,
    Data,
    EmptyInitFn,
    ForwardFn,
    ForwardInitFn,
    PartialBackFn,
    Shape,
    global_dtype,
)
from .utils import progress
from .utils.model_utils import to_ragged_seq_set
from .utils.parallel import clean_tempfile, memmap_buffer
from .utils.validation import check_vector


def _init_with_sequences(node, X, Y=None):
    """Initialize a Node with a sequence of inputs/targets."""
    X = to_ragged_seq_set(X)

    if Y is not None:
        Y = to_ragged_seq_set(Y)
    else:
        Y = [None for _ in range(len(X))]

    if not node.is_initialized:
        node.initialize(X[0], Y[0])

    return X, Y


def _init_vectors_placeholders(node, x, y):
    msg = f"Impossible to initialize node {node.name}: "
    in_msg = (
        msg + "input_dim is unknown and no input data x was given "
        "to call/run the node."
    )

    x_init = y_init = None
    if isinstance(x, np.ndarray):
        x_init = np.atleast_2d(check_vector(x, caller=node))
    elif isinstance(x, list):
        x_init = list()
        for i in range(len(x)):
            x_init.append(np.atleast_2d(check_vector(x[i], caller=node)))
    elif x is None:
        if node.input_dim is not None:
            if hasattr(node.input_dim, "__iter__"):
                x_init = [np.empty((1, d)) for d in node.input_dim]
            else:
                x_init = np.empty((1, node.input_dim))
        else:
            raise RuntimeError(in_msg)

    if y is not None:
        y_init = np.atleast_2d(check_vector(y, caller=node))
    elif node.output_dim is not None:
        y_init = np.empty((1, node.output_dim))
    else:
        # check if output dimension can be inferred from a teacher node
        if node._teacher is not None and node._teacher.output_dim is not None:
            y_init = np.empty((1, node._teacher.output_dim))

    return x_init, y_init


def _partial_backward_default(node, X_batch, Y_batch=None):
    """By default, for offline learners, partial_fit simply stores inputs and
    targets, waiting for fit to be called."""

    node._X.append(X_batch)

    if Y_batch is not None:
        node._Y.append(Y_batch)

    return


def _initialize_feedback_default(node, fb):
    """Void feedback initializer. Works in any case."""
    fb_dim = None
    if isinstance(fb, list):
        fb_dim = tuple([fb.shape[1] for fb in fb])
    elif isinstance(fb, np.ndarray):
        fb_dim = fb.shape[1]

    node.set_feedback_dim(fb_dim)


def _filter_where_na_target(x, y):

    if x.ndim == 1 and y.ndim == 1:
        is_na = np.isnan(y)
    elif x.ndim == 2 and y.ndim == 2:
        is_na = np.any(np.isnan(y), axis=-1)
    else:
        raise ValueError(
            "The dimensions of X_batch and Y_batch do not fit the expected cases."
        )

    is_kept = np.logical_not(is_na)
    return x[is_kept], y[is_kept]


class Node(_Node):
    """Node base class.

    Parameters
    ----------
    params : dict, optional
        Parameters of the Node. Parameters are mutable, and can be modified
        through learning or by the effect of hyperparameters.
    hypers : dict, optional
        Hyperparameters of the Node. Hyperparameters are immutable, and define
        the architecture and properties of the Node.
    forward : callable, optional
        A function defining the computation performed by the Node on some data
        point :math:`x_t`, and that would update the Node internal state from
        :math:`s_t` to :math:`s_{t+1}`.
    backward : callable, optional
        A function defining an offline learning rule, applied on a whole
        dataset, or on pre-computed values stored in buffers.
    partial_backward : callable, optional
        A function defining an offline learning rule, applied on a single batch
        of data.
    train : callable, optional
        A function defining an online learning, applied on a single step of
        a sequence or of a timeseries.
    initializer : callable, optional
        A function called at first run of the Node, defining the dimensions and
        values of its parameters based on the dimension of input data and its
        hyperparameters.
    fb_initializer : callable, optional
        A function called at first run of the Node, defining the dimensions and
        values of its parameters based on the dimension of data received as
        a feedback from another Node.
    buffers_initializer : callable, optional
        A function called at the beginning of an offline training session to
        create buffers used to store intermediate results, for batch or
        multi-sequence offline learning.
    input_dim : int
        Input dimension of the Node.
    output_dim : int
        Output dimension of the Node. Dimension of its state.
    feedback_dim :
        Dimension of the feedback signal received by the Node.
    name : str
        Name of the Node. It must be a unique identifier.

    See also
    --------
        Model
            Object used to compose node operations and create computational
            graphs.
    """

    _name: str

    _state: Optional[np.ndarray]
    _state_proxy: Optional[np.ndarray]
    _feedback: Optional[DistantFeedback]
    _teacher: Optional[DistantFeedback]

    _params: Dict[str, Any]
    _hypers: Dict[str, Any]
    _buffers: Dict[str, Any]

    _input_dim: int
    _output_dim: int
    _feedback_dim: int

    _forward: ForwardFn
    _backward: BackwardFn
    _partial_backward: PartialBackFn
    _train: PartialBackFn

    _initializer: ForwardInitFn
    _buffers_initializer: EmptyInitFn
    _feedback_initializer: ForwardInitFn

    _trainable: bool
    _fitted: bool

    _X: List  # For partial_fit default behavior (store first, then fit)
    _Y: List

    def __init__(
        self,
        params: Dict[str, Any] = None,
        hypers: Dict[str, Any] = None,
        forward: ForwardFn = None,
        backward: BackwardFn = None,
        partial_backward: PartialBackFn = _partial_backward_default,
        train: PartialBackFn = None,
        initializer: ForwardInitFn = None,
        fb_initializer: ForwardInitFn = _initialize_feedback_default,
        buffers_initializer: EmptyInitFn = None,
        input_dim: int = None,
        output_dim: int = None,
        feedback_dim: int = None,
        name: str = None,
        dtype: np.dtype = global_dtype,
    ):
        self._params = dict() if params is None else params
        self._hypers = dict() if hypers is None else hypers
        # buffers are all node state components that should not live
        # outside the node training loop, like partial computations for
        # linear regressions. They can also be shared across multiple processes
        # when needed.
        self._buffers = dict()

        self._forward = forward
        self._backward = backward
        self._partial_backward = partial_backward
        self._train = train

        self._initializer = initializer
        self._feedback_initializer = fb_initializer
        self._buffers_initializer = buffers_initializer

        self._input_dim = input_dim
        self._output_dim = output_dim
        self._feedback_dim = feedback_dim

        self._name = self._get_name(name)
        self._dtype = dtype

        self._is_initialized = False
        self._is_fb_initialized = False
        self._state_proxy = None
        self._feedback = None
        self._teacher = None
        self._fb_flag = True  # flag is used to trigger distant feedback model update

        self._trainable = self._backward is not None or self._train is not None

        self._fitted = False if self.is_trainable and self.is_trained_offline else True

        self._X, self._Y = [], []

    def __lshift__(self, other) -> "_Node":
        return self.link_feedback(other)

    def __ilshift__(self, other) -> "_Node":
        return self.link_feedback(other, inplace=True)

    def __iand__(self, other):
        raise TypeError(
            f"Impossible to merge nodes in-place: {self} is not a Model instance."
        )

    def _flag_feedback(self):
        self._fb_flag = not self._fb_flag

    def _unregister_teacher(self):
        self._teacher = None

    @property
    def input_dim(self):
        """Node input dimension."""
        return self._input_dim

    @property
    def output_dim(self):
        """Node output and internal state dimension."""
        return self._output_dim

    @property
    def feedback_dim(self):
        """Node feedback signal dimension."""
        return self._feedback_dim

    @property
    def is_initialized(self):
        """Returns if the Node is initialized or not."""
        return self._is_initialized

    @property
    def has_feedback(self):
        """Returns if the Node receives feedback or not."""
        return self._feedback is not None

    @property
    def is_trained_offline(self):
        """Returns if the Node can be fitted offline or not."""
        return self.is_trainable and self._backward is not None

    @property
    def is_trained_online(self):
        """Returns if the Node can be trained online or not."""
        return self.is_trainable and self._train is not None

    @property
    def is_trainable(self):
        """Returns if the Node can be trained."""
        return self._trainable

    @is_trainable.setter
    def is_trainable(self, value: bool):
        """Freeze or unfreeze the Node. If set to False,
        learning is stopped."""
        if self.is_trained_offline or self.is_trained_online:
            if type(value) is bool:
                self._trainable = value
            else:
                raise TypeError("'is_trainable' must be a boolean.")

    @property
    def fitted(self):
        """Returns if the Node parameters have fitted already, using an
        offline learning rule. If the node is trained online, returns True."""
        return self._fitted

    @property
    def is_fb_initialized(self):
        """Returns if the Node feedback initializer has been called already."""
        return self._is_fb_initialized

    @property
    def dtype(self):
        """Numpy numerical type of node parameters."""
        return self._dtype

    @property
    def unsupervised(self):
        return False

    def state(self) -> Optional[np.ndarray]:
        """Node current internal state.

        Returns
        -------
        array of shape (1, output_dim), optional
            Internal state of the Node.
        """
        if not self.is_initialized:
            return None
        return self._state

    def state_proxy(self) -> Optional[np.ndarray]:
        """Returns the internal state frozen to be sent to other Nodes,
        connected through a feedback connection. This prevents any change
        occurring on the Node before feedback have reached the other Node to
        propagate to the other Node to early.

        Returns
        -------
        array of shape (1, output_dim), optional
            Internal state of the Node.
        """
        if self._state_proxy is None:
            return self._state
        return self._state_proxy

    def feedback(self) -> np.ndarray:
        """State of the Nodes connected to this Node through feedback
        connections.

        Returns
        -------
        array-like of shape ([n_feedbacks], 1, feedback_dim), optional
            State of the feedback Nodes, i.e. the feedback signal.
        """
        if self.has_feedback:
            return self._feedback()
        else:
            raise RuntimeError(
                f"Node {self} is not connected to any feedback Node or Model."
            )

    def set_state_proxy(self, value: np.ndarray = None):
        """Change the frozen state of the Node. Used internally to send
        the current state to feedback receiver Nodes during the next call.

        Parameters
        ----------
        value : array of shape (1, output_dim)
            State to freeze, waiting to be sent to feedback receivers.
        """
        if value is not None:
            if self.is_initialized:
                value = check_one_sequence(
                    value, self.output_dim, allow_timespans=False, caller=self
                ).astype(self.dtype)
                self._state_proxy = value
            else:
                raise RuntimeError(f"{self.name} is not initialized yet.")

    def set_input_dim(self, value: int):
        """Set the input dimension of the Node. Can only be called once,
        during Node initialization."""
        if not self._is_initialized:
            if self._input_dim is not None and value != self._input_dim:
                raise ValueError(
                    f"Impossible to use {self.name} with input "
                    f"data of dimension {value}. Node has input "
                    f"dimension {self._input_dim}."
                )
            self._input_dim = value
        else:
            raise TypeError(
                f"Input dimension of {self.name} is immutable after initialization."
            )

    def set_output_dim(self, value: int):
        """Set the output dimension of the Node. Can only be called once,
        during Node initialization."""
        if not self._is_initialized:
            if self._output_dim is not None and value != self._output_dim:
                raise ValueError(
                    f"Impossible to use {self.name} with target "
                    f"data of dimension {value}. Node has output "
                    f"dimension {self._output_dim}."
                )
            self._output_dim = value
        else:
            raise TypeError(
                f"Output dimension of {self.name} is immutable after initialization."
            )

    def set_feedback_dim(self, value: int):
        """Set the feedback dimension of the Node. Can only be called once,
        during Node initialization."""
        if not self.is_fb_initialized:
            self._feedback_dim = value
        else:
            raise TypeError(
                f"Output dimension of {self.name} is immutable after initialization."
            )

    def get_param(self, name: str):
        """Get one of the parameters or hyperparameters given its name."""
        if name in self._params:
            return self._params.get(name)
        elif name in self._hypers:
            return self._hypers.get(name)
        else:
            raise AttributeError(f"No attribute named '{name}' found in node {self}")

    def set_param(self, name: str, value: Any):
        """Set the value of a parameter.

        Parameters
        ----------
        name : str
            Parameter name.
        value : array-like
            Parameter new value.
        """
        if name in self._params:
            if hasattr(value, "dtype"):
                if issparse(value):
                    value.data = value.data.astype(self.dtype)
                else:
                    value = value.astype(self.dtype)
            self._params[name] = value
        elif name in self._hypers:
            self._hypers[name] = value
        else:
            raise KeyError(
                f"No param named '{name}' "
                f"in {self.name}. Available params are: "
                f"{list(self._params.keys())}."
            )

    def create_buffer(
        self, name: str, shape: Shape = None, data: np.ndarray = None, as_memmap=True
    ):
        """Create a buffer array on disk, using numpy.memmap. This can be
        used to store transient variables on disk. Typically, called inside
        a `buffers_initializer` function.

        Parameters
        ----------
        name : str
            Name of the buffer array.
        shape : tuple of int, optional
            Shape of the buffer array.
        data : array-like
            Data to store in the buffer array.
        """
        if as_memmap:
            self._buffers[name] = memmap_buffer(self, data=data, shape=shape, name=name)
        else:
            if data is not None:
                self._buffers[name] = data
            else:
                self._buffers[name] = np.empty(shape)

    def set_buffer(self, name: str, value: np.ndarray):
        """Dump data in the buffer array.

        Parameters
        ----------
        name : str
            Name of the buffer array.
        value : array-like
            Data to store in the buffer array.
        """
        self._buffers[name][:] = value.astype(self.dtype)

    def get_buffer(self, name) -> np.memmap:
        """Get data from a buffer array.

        Parameters
        ----------
        name : str
            Name of the buffer array.

        Returns
        -------
            numpy.memmap
                Data as Numpy memory map.
        """
        if self._buffers.get(name) is None:
            raise AttributeError(f"No buffer named '{name}' in {self}.")
        return self._buffers[name]

    def initialize(self, x: Data = None, y: Data = None) -> "Node":
        """Call the Node initializers on some data points.
        Initializers are functions called at first run of the Node,
        defining the dimensions and values of its parameters based on the
        dimension of some input data and its hyperparameters.

        Data point `x` is used to infer the input dimension of the Node.
        Data point `y` is used to infer the output dimension of the Node.

        Parameters
        ----------
        x : array-like of shape ([n_inputs], 1, input_dim)
            Input data.
        y : array-like of shape (1, output_dim)
            Ground truth data. Used to infer output dimension
            of trainable nodes.

        Returns
        -------
        Node
            Initialized Node.
        """
        if not self.is_initialized:
            x_init, y_init = _init_vectors_placeholders(self, x, y)
            self._initializer(self, x=x_init, y=y_init)
            self.reset()
            self._is_initialized = True
        return self

    def initialize_feedback(self) -> "Node":
        """Call the Node feedback initializer. The feedback initializer will
        determine feedback dimension given some feedback signal, and initialize
        all parameters related to the feedback connection.

        Feedback sender Node must be initialized, as the feedback initializer
        will probably call the :py:meth:`Node.feedback` method to get
        a sample of feedback signal.

        Returns
        -------
        Node
            Initialized Node.
        """
        if self.has_feedback:
            if not self.is_fb_initialized:
                self._feedback.initialize()
                self._feedback_initializer(self, self.zero_feedback())
                self._is_fb_initialized = True
        return self

    def initialize_buffers(self) -> "Node":
        """Call the Node buffer initializer. The buffer initializer will create
        buffer array on demand to store transient values of the parameters,
        typically during training.

        Returns
        -------
        Node
            Initialized Node.
        """
        if self._buffers_initializer is not None:
            if len(self._buffers) == 0:
                self._buffers_initializer(self)

        return self

    def clean_buffers(self):
        """Clean Node's buffer arrays."""
        if len(self._buffers) > 0:
            self._buffers = dict()
            clean_tempfile(self)

        # Empty possibly stored inputs and targets in default buffer.
        self._X = self._Y = []

    def reset(self, to_state: np.ndarray = None) -> "Node":
        """Reset the last state saved to zero or to
        another state value `to_state`.

        Parameters
        ----------
        to_state : array of shape (1, output_dim), optional
            New state value.

        Returns
        -------
        Node
            Reset Node.
        """
        if to_state is None:
            self._state = self.zero_state()
        else:
            self._state = check_one_sequence(
                to_state, self.output_dim, allow_timespans=False, caller=self
            ).astype(self.dtype)
        return self

    @contextmanager
    def with_state(
        self, state: np.ndarray = None, stateful: bool = False, reset: bool = False
    ) -> "Node":
        """Modify the state of the Node using a context manager.
        The modification will have effect only within the context defined,
        before the state returns back to its previous value.

        Parameters
        ----------
        state : array of shape (1, output_dim), optional
            New state value.
        stateful : bool, default to False
            If set to True, then all modifications made in the context manager
            will remain after leaving the context.
        reset : bool, default to False
            If True, the Node will be reset using its :py:meth:`Node.reset`
            method.

        Returns
        -------
        Node
            Modified Node.
        """
        if not self._is_initialized:
            raise RuntimeError(
                f"Impossible to set state of node {self.name}: node"
                f" is not initialized yet."
            )

        current_state = self._state

        if state is None:
            if reset:
                state = self.zero_state()
            else:
                state = current_state

        self.reset(to_state=state)
        yield self

        if not stateful:
            self._state = current_state

    @contextmanager
    def with_feedback(
        self, feedback: np.ndarray = None, stateful=False, reset=False
    ) -> "Node":
        """Modify the feedback received or sent by the Node using
        a context manager.
        The modification will have effect only within the context defined,
        before the feedback returns to its previous state.

        If the Node is receiving feedback, then this function will alter the
        state of the Node connected to it through feedback connections.

        If the Node is sending feedback, then this function will alter the
        state (or state proxy, see :py:meth:`Node.state_proxy`) of the Node.

        Parameters
        ----------
        feedback : array of shape (1, feedback_dim), optional
            New feedback signal.
        stateful : bool, default to False
            If set to True, then all modifications made in the context manager
            will remain after leaving the context.
        reset : bool, default to False
            If True, the feedback  will be reset to zero.

        Returns
        -------
        Node
            Modified Node.
        """
        if self.has_feedback:
            if reset:
                feedback = self.zero_feedback()
            if feedback is not None:
                self._feedback.clamp(feedback)

            yield self

        else:  # maybe a feedback sender then?
            current_state_proxy = self._state_proxy

            if feedback is None:
                if reset:
                    feedback = self.zero_state()
                else:
                    feedback = current_state_proxy

            self.set_state_proxy(feedback)

            yield self

            if not stateful:
                self._state_proxy = current_state_proxy

    def zero_state(self) -> np.ndarray:
        """A null state vector."""
        if self.output_dim is not None:
            return np.zeros((1, self.output_dim), dtype=self.dtype)

    def zero_feedback(self) -> Optional[Union[List[np.ndarray], np.ndarray]]:
        """A null feedback vector. Returns None if the Node receives
        no feedback."""
        if self._feedback is not None:
            return self._feedback.zero_feedback()
        return None

    def link_feedback(
        self, node: _Node, inplace: bool = False, name: str = None
    ) -> "_Node":
        """Create a feedback connection between the Node and another Node or
        Model.

        Parameters
        ----------
        node : Node or Model
            Feedback sender Node or Model.
        inplace : bool, default to False
            If False, then this function returns a copy of the current Node
            with feedback enabled. If True, feedback is directly added to the
            current Node.
        name : str, optional
            Name of the node copy, if `inplace` is False.

        Returns
        -------
        Node
            A Node with a feedback connection.
        """
        from .ops import link_feedback

        return link_feedback(self, node, inplace=inplace, name=name)

    def call(
        self,
        x: Data,
        from_state: np.ndarray = None,
        stateful: bool = True,
        reset: bool = False,
    ) -> np.ndarray:
        """Call the Node forward function on a single step of data.
        Can update the state of the
        Node.

        Parameters
        ----------
        x : array of shape ([n_inputs], 1, input_dim)
            One single step of input data.
        from_state : array of shape (1, output_dim), optional
            Node state value to use at beginning of computation.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Node state will be reset to zero before this operation.

        Returns
        -------
        array of shape (1, output_dim)
            An output vector.
        """
        x, _ = check_xy(
            self,
            x,
            allow_timespans=False,
            allow_n_sequences=False,
        )

        if not self._is_initialized:
            self.initialize(x)

        return call(self, x, from_state=from_state, stateful=stateful, reset=reset)

    def run(self, X: np.array, from_state=None, stateful=True, reset=False):
        """Run the Node forward function on a sequence of data.
        Can update the state of the
        Node several times.

        Parameters
        ----------
        X : array-like of shape ([n_inputs], timesteps, input_dim)
            A sequence of data of shape (timesteps, features).
        from_state : array of shape (1, output_dim), optional
            Node state value to use at beginning of computation.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Node state will be reset to zero before this operation.

        Returns
        -------
        array of shape (timesteps, output_dim)
            A sequence of output vectors.
        """
        X_, _ = check_xy(
            self,
            X,
            allow_n_sequences=False,
        )

        if isinstance(X_, np.ndarray):
            if not self._is_initialized:
                self.initialize(np.atleast_2d(X_[0]))
            seq_len = X_.shape[0]
        else:  # multiple inputs?
            if not self._is_initialized:
                self.initialize([np.atleast_2d(x[0]) for x in X_])
            seq_len = X_[0].shape[0]

        with self.with_state(from_state, stateful=stateful, reset=reset):
            states = np.zeros((seq_len, self.output_dim))
            for i in progress(range(seq_len), f"Running {self.name}: "):
                if isinstance(X_, (list, tuple)):
                    x = [np.atleast_2d(Xi[i]) for Xi in X_]
                else:
                    x = np.atleast_2d(X_[i])

                s = call(self, x)
                states[i, :] = s

        return states

    def train(
        self,
        X: np.ndarray,
        Y: Union[_Node, np.ndarray] = None,
        force_teachers: bool = True,
        call: bool = True,
        learn_every: int = 1,
        from_state: np.ndarray = None,
        stateful: bool = True,
        reset: bool = False,
    ) -> np.ndarray:
        """Train the Node parameters using an online learning rule, if
        available.

        Parameters
        ----------
        X : array-like of shape ([n_inputs], timesteps, input_dim)
            Input sequence of data.
        Y : array-like of shape (timesteps, output_dim), optional.
            Target sequence of data. If None, the Node will search a feedback
            signal, or train in an unsupervised way, if possible.
        force_teachers : bool, default to True
            If True, this Node will broadcast the available ground truth signal
            to all Nodes using this Node as a feedback sender. Otherwise,
            the real state of this Node will be sent to the feedback receivers.
        call : bool, default to True
            It True, call the Node and update its state before applying the
            learning rule. Otherwise, use the train method
            on the current state.
        learn_every : int, default to 1
            Time interval at which training must occur, when dealing with a
            sequence of input data. By default, the training method is called
            every time the Node receive an input.
        from_state : array of shape (1, output_dim), optional
            Node state value to use at beginning of computation.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Node state will be reset to zero before this operation.

        Returns
        -------
        array of shape (timesteps, output_dim)
            All outputs computed during the training. If `call` is False,
            outputs will be the result of :py:meth:`Node.zero_state`.
        """
        if not self.is_trained_online:
            raise TypeError(f"Node {self} has no online learning rule implemented.")

        X_, Y_ = check_xy(
            self,
            X,
            Y,
            allow_n_sequences=False,
            allow_n_inputs=False,
        )

        if not self._is_initialized:
            x_init = np.atleast_2d(X_[0])
            y_init = None
            if hasattr(Y, "__iter__"):
                y_init = np.atleast_2d(Y_[0])

            self.initialize(x=x_init, y=y_init)
            self.initialize_buffers()

        states = train(
            self,
            X_,
            Y_,
            call_node=call,
            force_teachers=force_teachers,
            learn_every=learn_every,
            from_state=from_state,
            stateful=stateful,
            reset=reset,
        )

        self._unregister_teacher()

        return states

    def partial_fit(
        self,
        X_batch: Data,
        Y_batch: Data = None,
        warmup=0,
        **kwargs,
    ) -> "Node":
        """Partial offline fitting method of a Node.
        Can be used to perform batched fitting or to pre-compute some variables
        used by the fitting method.

        Parameters
        ----------
        X_batch : array-like of shape ([n_inputs], [series], timesteps, input_dim)
            A sequence or a batch of sequence of input data.
        Y_batch : array-like of shape ([series], timesteps, output_dim), optional
            A sequence or a batch of sequence of teacher signals.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.

        Returns
        -------
        Node
            Partially fitted Node.
        """
        if not self.is_trained_offline:
            raise TypeError(f"Node {self} has no offline learning rule implemented.")

        X, Y = check_xy(self, X_batch, Y_batch, allow_n_inputs=False)

        X, Y = _init_with_sequences(self, X, Y)

        self.initialize_buffers()

        for i in range(len(X)):
            X_seq = X[i]
            Y_seq = None
            if Y is not None:
                Y_seq = Y[i]

            if X_seq.shape[0] <= warmup:
                raise ValueError(
                    f"Warmup set to {warmup} timesteps, but one timeseries is only "
                    f"{X_seq.shape[0]} long."
                )

            if Y_seq is not None:
                X_nona, Y_nona = _filter_where_na_target(X_seq[warmup:], Y_seq[warmup:])
                self._partial_backward(self, X_nona, Y_nona, **kwargs)
            else:
                self._partial_backward(self, X_seq[warmup:], **kwargs)

        return self

    def fit(self, X: Data = None, Y: Data = None, warmup=0) -> "Node":
        """Offline fitting method of a Node.

        Parameters
        ----------
        X : array-like of shape ([n_inputs], [series], timesteps, input_dim), optional
            Input sequences dataset. If None, the method will try to fit
            the parameters of the Node using the precomputed values returned
            by previous call of :py:meth:`partial_fit`.
        Y : array-like of shape ([series], timesteps, output_dim), optional
            Teacher signals dataset. If None, the method will try to fit
            the parameters of the Node using the precomputed values returned
            by previous call of :py:meth:`partial_fit`, or to fit the Node in
            an unsupervised way, if possible.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.

        Returns
        -------
        Node
            Node trained offline.
        """

        if not self.is_trained_offline:
            raise TypeError(f"Node {self} has no offline learning rule implemented.")

        self._fitted = False

        # Call the partial backward function on the dataset if it is
        # provided all at once.
        if X is not None:
            if self._partial_backward is not None:
                self.partial_fit(X, Y, warmup=warmup)

        elif not self._is_initialized:
            raise RuntimeError(
                f"Impossible to fit node {self.name}: node"
                f" is not initialized, and fit was called "
                f"without input and teacher data."
            )

        self._backward(self, self._X, self._Y)

        self._fitted = True
        self.clean_buffers()

        return self

    def copy(
        self, name: str = None, copy_feedback: bool = False, shallow: bool = False
    ):
        """Returns a copy of the Node.

        Parameters
        ----------
        name : str
            Name of the Node copy.
        copy_feedback : bool, default to False
            If True, also copy the Node feedback senders.
        shallow : bool, default to False
            If False, performs a deep copy of the Node.

        Returns
        -------
        Node
            A copy of the Node.
        """
        if shallow:
            new_obj = copy(self)
        else:
            if self.has_feedback:
                # store feedback node
                fb = self._feedback
                # temporarily remove it
                self._feedback = None

                # copy and restore feedback, deep copy of feedback depends
                # on the copy_feedback parameter only
                new_obj = deepcopy(self)
                new_obj._feedback = fb
                self._feedback = fb

            else:
                new_obj = deepcopy(self)

        if copy_feedback:
            if self.has_feedback:
                fb_copy = deepcopy(self._feedback)
                new_obj._feedback = fb_copy

        n = self._get_name(name)
        new_obj._name = n

        return new_obj


class Unsupervised(Node):
    @property
    def unsupervised(self):
        return True



"""
===================================
Models (:class:`reservoirpy.Model`)
===================================

Note
----

See the following guides to:

- **Learn more about how to work with ReservoirPy Nodes**: :ref:`node`
- **Learn more about how to combine nodes within a Model**: :ref:`model`


Models are an extension of the Node API. They allow to combine nodes into
complex computational graphs, to create complicated Reservoir Computing
architecture like *Deep Echo State Networks*.

See :ref:`model` to learn more about how to create and manipulate
a :py:class:`Model`.

.. currentmodule:: reservoirpy.model

.. autoclass:: Model

   .. rubric:: Methods

   .. autosummary::

      ~Model.call
      ~Model.fit
      ~Model.get_node
      ~Model.initialize
      ~Model.initialize_buffers
      ~Model.reset
      ~Model.run
      ~Model.train
      ~Model.update_graph
      ~Model.with_feedback
      ~Model.with_state


   .. rubric:: Attributes

   .. autosummary::

      ~Model.data_dispatcher
      ~Model.edges
      ~Model.feedback_nodes
      ~Model.fitted
      ~Model.hypers
      ~Model.input_dim
      ~Model.input_nodes
      ~Model.is_empty
      ~Model.is_initialized
      ~Model.is_trainable
      ~Model.is_trained_offline
      ~Model.is_trained_online
      ~Model.name
      ~Model.node_names
      ~Model.nodes
      ~Model.output_dim
      ~Model.output_nodes
      ~Model.params
      ~Model.trainable_nodes

.. autoclass:: FrozenModel

"""
# Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from functools import partial
from typing import Dict, List, Optional, Sequence, Tuple

import numpy as np

from . import _base
from ._base import _Node, check_xy
from .type import MappedData
from .utils import progress, verbosity
from .utils.graphflow import (
    DataDispatcher,
    dispatch,
    find_entries_and_exits,
    get_offline_subgraphs,
    topological_sort,
)
from .utils.model_utils import (
    allocate_returned_states,
    build_forward_sumodels,
    dist_states_to_next_subgraph,
    fold_mapping,
    to_data_mapping,
)
from .utils.validation import is_mapping


def run_and_partial_fit(
    submodel,
    complete_model,
    offlines,
    relations,
    x_seq,
    y_seq,
    warmup,
    stateful=True,
    reset=False,
    return_states=None,
    force_teachers=True,
):
    """Run a submodel and call its partial fit function."""

    if not submodel.is_empty:
        x_seq = {n: x for n, x in x_seq.items() if n in submodel.node_names}

        if force_teachers and y_seq is not None:
            y_seq = {
                n: y
                for n, y in y_seq.items()
                if n in [o.name for o in complete_model.nodes if o.is_trained_offline]
            }
        else:
            y_seq = None

        submodel._is_initialized = True
        states = run_submodel(
            complete_model,
            submodel,
            x_seq,
            y_seq,
            return_states=return_states,
            stateful=stateful,
            reset=reset,
        )

        dist_states = dist_states_to_next_subgraph(states, relations)
    else:
        dist_states = x_seq

    if y_seq is not None:
        for node in offlines:
            node.partial_fit(
                dist_states.get(node.name), y_seq.get(node.name), warmup=warmup
            )
    else:
        for node in offlines:
            node.partial_fit(dist_states.get(node.name), warmup=warmup)

    return dist_states


def run_submodel(
    model,
    submodel,
    X: MappedData,
    forced_feedbacks: Dict[str, np.ndarray] = None,
    from_state: Dict[str, np.ndarray] = None,
    stateful=True,
    reset=False,
    shift_fb=True,
    return_states: Sequence[str] = None,
) -> MappedData:

    X_, forced_feedbacks_ = to_data_mapping(submodel, X, forced_feedbacks)

    submodel._initialize_on_sequence(X_[0], forced_feedbacks_[0])

    states = []
    for X_seq, fb_seq in zip(X_, forced_feedbacks_):
        with model.with_state(reset=reset, stateful=stateful):
            states_seq = model._run(
                X_seq,
                fb_seq,
                from_state=from_state,
                stateful=stateful,
                shift_fb=shift_fb,
                return_states=return_states,
                submodel=submodel,
            )

            states.append(states_seq)

    return fold_mapping(submodel, states, return_states)


def forward(model: "Model", x: MappedData) -> List[np.ndarray]:
    """Function applied by a :py:class:`Model` instance.

    This function if basically a composition of the forward functions of all
    nodes involved in the model architecture. For instance, let :math:`f`
    be the forward function of a first node, and let :math:`g` be the forward
    function of a second node. If first node is connected to second node in a
    model, then the model forward function :math:`h` will compute, at each
    timestep :math:`t` of a timeseries :math:`X`:

    .. math::

        h(X_t) = g(f(X_t)) = (g \\circ f)(X_t)

    Parameters
    ----------
    model : Model
        A :py:class:`Model` instance.
    x : dict or array-like of shape ([n_inputs], 1, input_dim)
        A vector corresponding to a timestep of data, or
        a dictionary mapping node names to input vectors.

    Returns
    -------
    array-like of shape (n_outputs, 1, output_dim)
        New states of all terminal nodes of the model, i.e. its output.
    """
    data = model.data_dispatcher.load(x)

    for node in model.nodes:
        _base.call(node, data[node].x)

    return [out_node.state() for out_node in model.output_nodes]


def train(model: "Model", x=None, y: MappedData = None, force_teachers=True):
    """Training function for a Model. Run all train functions of all online
    nodes within the Model. Nodes have already been called. Only training
    is performed."""

    data = model.data_dispatcher.load(X=x, Y=y)

    for node in model.nodes:
        if node.is_trained_online:
            _base.train(
                node,
                data[node].x,
                data[node].y,
                force_teachers=force_teachers,
                call_node=False,
            )


def initializer(model: "Model", x: MappedData, y: Optional[MappedData] = None):
    """Initializes a :py:class:`Model` instance at runtime, using samples of
    data to infer all :py:class:`Node` dimensions.

    Parameters
    ----------
    model : Model
        A :py:class:`Model` instance.
    x : numpy.ndarray or dict of numpy.ndarray
        A vector of shape `(1, ndim)` corresponding to a timestep of data, or
        a dictionary mapping node names to vector of shapes
        `(1, ndim of corresponding node)`.
    y : numpy.ndarray or dict of numpy.ndarray
        A vector of shape `(1, ndim)` corresponding to a timestep of target
        data or feedback, or a dictionary mapping node names to vector of
        shapes `(1, ndim of corresponding node)`.
    """
    data = model.data_dispatcher.load(x, y)

    # first, probe network to init forward flow
    # (no real call, only zero states)
    for node in model.nodes:
        node.initialize(x=data[node].x, y=data[node].y)

    # second, probe feedback demanding nodes to
    # init feedback flow
    for fb_node in model.feedback_nodes:
        fb_node.initialize_feedback()


class Model(_Node):
    """Model base class.

    Parameters
    ----------
    nodes : list of Node, optional
        Nodes to include in the Model.
    edges : list of (Node, Node), optional
        Edges between Nodes in the graph. An edge between a
        Node A and a Node B is created as a tuple (A, B).
    name : str, optional
        Name of the Model.
    """

    _node_registry: Dict[str, _Node]
    _nodes: List[_Node]
    _edges: List[Tuple[_Node, _Node]]
    _inputs: List[_Node]
    _outputs: List[_Node]
    _dispatcher: "DataDispatcher"

    def __init__(
        self,
        nodes: Sequence[_Node] = None,
        edges: Sequence[Tuple[_Node, _Node]] = None,
        name: str = None,
    ):

        if nodes is None:
            nodes = tuple()
        if edges is None:
            edges = tuple()

        self._forward = forward
        self._train = train
        self._initializer = initializer

        self._name = self._get_name(name)

        nodes, edges = self._concat_multi_inputs(nodes, edges)

        self._edges = edges

        # always maintain nodes in topological order
        if len(nodes) > 0:
            self._inputs, self._outputs = find_entries_and_exits(nodes, edges)
            self._nodes = topological_sort(nodes, edges, self._inputs)
        else:
            self._inputs = list()
            self._outputs = list()
            self._nodes = nodes

        self._is_initialized = False
        self._trainable = any([n.is_trainable for n in nodes])
        self._fitted = all([n.fitted for n in nodes])

        self._params = {n.name: n.params for n in nodes}
        self._hypers = {n.name: n.hypers for n in nodes}

        self._node_registry = {n.name: n for n in self.nodes}

        self._dispatcher = DataDispatcher(self)

    def __repr__(self):
        klas = self.__class__.__name__
        nodes = [n.name for n in self._nodes]
        return f"'{self.name}': {klas}('" + "', '".join(nodes) + "')"

    def __getitem__(self, item):
        return self.get_node(item)

    def __iand__(self, other) -> "Model":
        from .ops import merge

        return merge(self, other, inplace=True)

    @staticmethod
    def _concat_multi_inputs(nodes, edges):
        from .ops import concat_multi_inputs

        return concat_multi_inputs(nodes, edges)

    def _check_if_only_online(self):
        if any([n.is_trained_offline and not n.fitted for n in self.nodes]):
            raise RuntimeError(
                f"Impossible to train model {self.name} using "
                f"online method: model contains untrained "
                f"offline nodes."
            )

    def _load_proxys(self, keep=False):
        """Save states of all nodes into their state_proxy"""
        for node in self._nodes:
            if keep and node._state_proxy is not None:
                continue
            node._state_proxy = node.state()

    def _clean_proxys(self):
        """Destroy state_proxy of all nodes"""
        for node in self._nodes:
            node._state_proxy = None

    def _initialize_on_sequence(self, X=None, Y=None):
        if not self._is_initialized:
            x_init = None
            if X is not None:
                if is_mapping(X):
                    x_init = {name: np.atleast_2d(x[0]) for name, x in X.items()}
                else:
                    x_init = np.atleast_2d(X[0])

            y_init = None
            if Y is not None:
                if is_mapping(Y):
                    y_init = {name: np.atleast_2d(y[0]) for name, y in Y.items()}
                else:
                    y_init = np.atleast_2d(Y[0])

            self.initialize(x_init, y_init)

    def _call(self, x=None, return_states=None, submodel=None, *args, **kwargs):

        if submodel is None:
            submodel = self

        self._forward(submodel, x)

        state = {}
        if return_states == "all":
            for node in submodel.nodes:
                state[node.name] = node.state()

        elif hasattr(return_states, "__iter__"):
            for name in return_states:
                state[name] = submodel[name].state()

        else:
            if len(submodel.output_nodes) > 1:
                for out_node in submodel.output_nodes:
                    state[out_node.name] = out_node.state()
            else:
                state = submodel.output_nodes[0].state()

        return state

    def _run(
        self,
        X,
        feedback,
        from_state=None,
        stateful=True,
        shift_fb=True,
        return_states=None,
        submodel=None,
    ):
        if submodel is None:
            submodel = self

        states = allocate_returned_states(submodel, X, return_states)

        seq = progress(
            dispatch(X, feedback, shift_fb=shift_fb),
            f"Running {self.name}",
            total=len(X),
        )

        with self.with_state(from_state, stateful=stateful):
            # load proxys after state update to make it have an
            # impact on feedback
            self._load_proxys(keep=True)
            for i, (x, forced_fb, _) in enumerate(seq):

                with self.with_feedback(forced_fb):
                    state = submodel._call(x, return_states=return_states)

                if is_mapping(state):
                    for name, value in state.items():
                        states[name][i, :] = value
                else:
                    states[submodel.output_nodes[0].name][i, :] = state

                self._load_proxys()

        self._clean_proxys()

        return states

    def _unregister_teachers(self):
        """Remove teacher nodes refs from student nodes."""
        for node in self.trainable_nodes:
            node._teacher = None

    def update_graph(
        self,
        new_nodes: Sequence[_Node],
        new_edges: Sequence[Tuple[_Node, _Node]],
    ) -> "Model":
        """Update current Model's with new nodes and edges, inplace (a copy
        is not performed).

        Parameters
        ----------
        new_nodes : list of Node
            New nodes.
        new_edges : list of (Node, Node)
            New edges between nodes.

        Returns
        -------
        Model
            The updated Model.
        """
        nodes = list(set(new_nodes) | set(self.nodes))
        edges = list(set(new_edges) | set(self.edges))

        nodes, edges = self._concat_multi_inputs(nodes, edges)

        self._nodes = nodes
        self._edges = edges

        self._params = {n.name: n.params for n in self._nodes}
        self._hypers = {n.name: n.hypers for n in self._nodes}

        self._inputs, self._outputs = find_entries_and_exits(self._nodes, self._edges)
        self._nodes = topological_sort(self._nodes, self._edges, self._inputs)
        self._node_registry = {n.name: n for n in self.nodes}

        self._dispatcher = DataDispatcher(self)

        self._fitted = all([n.fitted for n in self.nodes])
        self._is_initialized = False

        return self

    def get_node(self, name: str) -> _Node:
        """Get Node in Model, by name.

        Parameters
        ----------
        name : str
            Node name.

        Returns
        -------
        Node
            The requested Node.

        Raises
        ------
        KeyError
            Node not found.
        """
        if self._node_registry.get(name) is not None:
            return self._node_registry[name]
        else:
            raise KeyError(f"No node named '{name}' found in model {self.name}.")

    @property
    def nodes(self) -> List[_Node]:
        """Nodes in the Model, in topological order."""
        return self._nodes

    @property
    def node_names(self):
        """Names of all the Nodes in the Model."""
        return list(self._node_registry.keys())

    @property
    def edges(self):
        """All edges between Nodes, in the form (sender, receiver)."""
        return self._edges

    @property
    def input_dim(self):
        """Input dimension of the Model;
        input dimensions of all input Nodes."""
        if self.is_initialized:
            dims = [n.input_dim for n in self.input_nodes]
            if len(dims) == 0:
                return 0
            elif len(dims) < 2:
                return dims[0]
            return dims
        else:
            return None

    @property
    def output_dim(self):
        """Output dimension of the Model;
        output dimensions of all output Nodes."""
        if self.is_initialized:
            dims = [n.output_dim for n in self.output_nodes]
            if len(dims) == 0:
                return 0
            elif len(dims) < 2:
                return dims[0]
            return dims
        else:
            return None

    @property
    def input_nodes(self):
        """First Nodes in the graph held by the Model."""
        return self._inputs

    @property
    def output_nodes(self):
        """Last Nodes in the graph held by the Model."""
        return self._outputs

    @property
    def trainable_nodes(self):
        """Returns all offline and online
        trainable Nodes in the Model."""
        return [n for n in self.nodes if n.is_trainable]

    @property
    def feedback_nodes(self):
        """Returns all Nodes equipped with a feedback connection
        in the Model."""
        return [n for n in self.nodes if n.has_feedback]

    @property
    def data_dispatcher(self):
        """DataDispatcher object of the Model. Used to
        distribute data to Nodes when
        calling/running/fitting the Model."""
        return self._dispatcher

    @property
    def is_empty(self):
        """Returns True if the Model contains no Nodes."""
        return len(self.nodes) == 0

    @property
    def is_trainable(self) -> bool:
        """Returns True if at least one Node in the Model is trainable."""
        return any([n.is_trainable for n in self.nodes])

    @is_trainable.setter
    def is_trainable(self, value):
        """Freeze or unfreeze trainable Nodes in the Model."""
        trainables = [
            n for n in self.nodes if n.is_trained_offline or n.is_trained_online
        ]
        for node in trainables:
            node.is_trainable = value

    @property
    def is_trained_online(self) -> bool:
        """Returns True if all nodes are online learners."""
        return all([n.is_trained_online or n.fitted for n in self.nodes])

    @property
    def is_trained_offline(self) -> bool:
        """Returns True if all nodes are offline learners."""
        return all([n.is_trained_offline or n.fitted for n in self.nodes])

    @property
    def fitted(self) -> bool:
        """Returns True if all nodes are fitted."""
        return all([n.fitted for n in self.nodes])

    @contextmanager
    def with_state(
        self, state: Dict[str, np.ndarray] = None, stateful=False, reset=False
    ) -> "Model":
        """Modify the state of one or several Nodes in the Model
        using a context manager.
        The modification will have effect only within the context defined,
        before all states return to their previous value.

        Parameters
        ----------
        state : dict of arrays of shape (1, output_dim), optional
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        stateful : bool, default to False
            If set to True, then all modifications made in the context manager
            will remain after leaving the context.
        reset : bool, default to False
            If True, all Nodes will be reset using their :py:meth:`Node.reset`
            method.

        Returns
        -------
        Model
            Modified Model.
        """
        if state is None and not reset:
            current_state = None
            if not stateful:
                current_state = {n.name: n.state() for n in self.nodes}
            yield self
            if not stateful:
                self.reset(to_state=current_state)
        elif isinstance(state, np.ndarray):
            raise TypeError(
                f"Impossible to set state of {self.name} with "
                f"an array. State should be a dict mapping state "
                f"to a Node name within the model."
            )
        else:
            if state is None:
                state = {}

            with ExitStack() as stack:
                for node in self.nodes:
                    value = state.get(node.name)
                    stack.enter_context(
                        node.with_state(value, stateful=stateful, reset=reset)
                    )
                yield self

    @contextmanager
    def with_feedback(
        self, feedback: Dict[str, np.ndarray] = None, stateful=False, reset=False
    ) -> "Model":
        """Modify the feedback received or sent by Nodes in the Model using
        a context manager.
        The modification will have effect only within the context defined,
        before the feedbacks return to their previous states.

        If the Nodes are receiving feedback, then this function will alter the
        states of the Nodes connected to it through feedback connections.

        If the Nodes are sending feedback, then this function will alter the
        states (or state proxies, see :py:meth:`Node.state_proxy`) of the
        Nodes.

        Parameters
        ----------
        feedback : dict of arrays of shape (1, output_dim), optional
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray feedback vectors.
        stateful : bool, default to False
            If set to True, then all modifications made in the context manager
            will remain after leaving the context.
        reset : bool, default to False
            If True, all feedbacks  will be reset to zero.

        Returns
        -------
        Model
            Modified Model.
        """

        if feedback is None and not reset:
            yield self
            return

        elif feedback is not None:
            with ExitStack() as stack:
                for node in self.nodes:
                    value = feedback.get(node.name)
                    # maybe node does not send feedback but receives it?
                    if value is None and node.has_feedback:
                        value = feedback.get(node._feedback.name)
                    stack.enter_context(
                        node.with_feedback(value, stateful=stateful, reset=reset)
                    )

                yield self
        else:
            yield self
            return

    def reset(self, to_state: Dict[str, np.ndarray] = None):
        """Reset the last state saved to zero for all Nodes in
        the Model or to other state values.

        Parameters
        ----------
        to_state : dict of arrays of shape (1, output_dim), optional
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        """
        if to_state is None:
            for node in self.nodes:
                node.reset()
        else:
            for node_name, current_state in to_state.items():
                self[node_name].reset(to_state=current_state)
        return self

    def initialize(self, x=None, y=None) -> "Model":
        """Call the Model initializers on some data points.
        Model will be virtually run to infer shapes of all nodes given
        inputs and targets vectors.

        Parameters
        ----------
        x : dict or array-like of shape ([n_inputs], 1, input_dim)
            Input data.
        y : dict or array-like of shape ([n_outputs], 1, output_dim)
            Ground truth data. Used to infer output dimension
            of trainable nodes.

        Returns
        -------
        Model
            Initialized Model.
        """
        self._is_initialized = False
        self._initializer(self, x=x, y=y)
        self.reset()
        self._is_initialized = True
        return self

    def initialize_buffers(self) -> "Model":
        """Call all Nodes buffer initializers. Buffer initializers will create
        buffer arrays on demand to store transient values of the parameters,
        typically during training.

        Returns
        -------
        Model
            Initialized Model.
        """
        for node in self.nodes:
            if node._buffers_initializer is not None:
                node.initialize_buffers()

    def call(
        self,
        x: MappedData,
        forced_feedback: MappedData = None,
        from_state: Dict[str, np.ndarray] = None,
        stateful=True,
        reset=False,
        return_states: Sequence[str] = None,
    ) -> MappedData:
        """Call the Model forward function on a single step of data.
        Model forward function is a composition of all its Nodes forward
        functions.

        Can update the state its Nodes.

        Parameters
        ----------
        x : dict or array-like of shape ([n_inputs], 1, input_dim)
            One single step of input data. If dict, then
            pairs of keys and values, where keys are Model input
            nodes names and values are single steps of input data.
        forced_feedback: dict of arrays of shape ([n_feedbacks], 1, feedback_dim), optional
            Pairs of keys and values, where keys are Model nodes names and
            value are feedback vectors to force into the nodes.
        from_state : dict of arrays of shape ([nodes], 1, nodes.output_dim), optional
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Nodes states will be reset to zero before this operation.
        return_states: list of str, optional
            Names of Nodes from which to return states as output.

        Returns
        -------
        dict or array-like of shape ([n_outputs], 1, output_dim)
            An output vector or pairs of keys and values
            where keys are output nodes names and values
            are corresponding output vectors.
        """

        x, _ = check_xy(
            self,
            x,
            allow_timespans=False,
            allow_n_sequences=False,
        )

        if not self._is_initialized:
            self.initialize(x)

        with self.with_state(from_state, stateful=stateful, reset=reset):
            # load current states in proxys interfaces accessible
            # through feedback. These proxys are not updated during the
            # graph call.
            self._load_proxys(keep=True)

            # maybe load forced feedbacks in proxys?
            with self.with_feedback(forced_feedback, stateful=stateful, reset=reset):
                state = self._call(x, return_states)

        # wash states proxys
        self._clean_proxys()

        return state

    def run(
        self,
        X: MappedData,
        forced_feedbacks: Dict[str, np.ndarray] = None,
        from_state: Dict[str, np.ndarray] = None,
        stateful=True,
        reset=False,
        shift_fb=True,
        return_states: Sequence[str] = None,
    ) -> MappedData:
        """Run the Model forward function on a sequence of data.
        Model forward function is a composition of all its Nodes forward
        functions.
        Can update the state of the
        Nodes several times.

        Parameters
        ----------
        X : dict or array-like of shape ([n_inputs], timesteps, input_dim)
            A sequence of input data.
            If dict, then pairs of keys and values, where keys are Model input
            nodes names and values are sequence of input data.
        forced_feedbacks: dict of array-like of shape ([n_feedbacks], timesteps, feedback_dim)
            Pairs of keys and values, where keys are Model nodes names and
            value are sequences of feedback vectors to force into the nodes.
        from_state : dict of arrays of shape ([nodes], 1, nodes.output_dim)
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Nodes states will be reset to zero before this operation.
        shift_fb: bool, defaults to True
            If True, then forced feedbacks are fed to nodes with a
            one timestep delay. If forced feedbacks are a sequence
            of target vectors matching the sequence of input
            vectors, then this parameter should be set to True.
        return_states: list of str, optional
            Names of Nodes from which to return states as output.

        Returns
        -------
        dict or array-like of shape ([n_outputs], timesteps, output_dim)
            A sequence of output vectors or pairs of keys and values
            where keys are output nodes names and values
            are corresponding sequences of output vectors.
        """
        X_, forced_feedbacks_ = to_data_mapping(self, X, forced_feedbacks)

        self._initialize_on_sequence(X_[0], forced_feedbacks_[0])

        states = []
        for X_seq, fb_seq in zip(X_, forced_feedbacks_):
            with self.with_state(reset=reset, stateful=stateful):
                states_seq = self._run(
                    X_seq,
                    fb_seq,
                    from_state=from_state,
                    stateful=stateful,
                    shift_fb=shift_fb,
                    return_states=return_states,
                )

                states.append(states_seq)

        return fold_mapping(self, states, return_states)

    def train(
        self,
        X,
        Y=None,
        force_teachers=True,
        learn_every=1,
        from_state=None,
        stateful=True,
        reset=False,
        return_states=None,
    ) -> MappedData:
        """Train all online Nodes in the Model
        using their online learning rule.

        Parameters
        ----------
        X : dict or array-like of shape ([n_inputs], timesteps, input_dim)
            Input sequence of data. If dict, then pairs
            of keys and values, where keys are Model input
            nodes names and values are sequence of input data.
        Y : dict or array-like of shape ([onlines], timesteps, onlines.output_dim), optional.
            Target sequence of data.
            If dict, then pairs of keys and values, where keys are Model
            online trainable nodes names values are sequences
            of target data. If None, the Nodes will search a feedback
            signal, or train in an unsupervised way, if possible.
        force_teachers : bool, default to True
            If True, this Model will broadcast the available ground truth
            signal
            to all online nodes sending feedback to other nodes. Otherwise,
            the real state of these nodes will be sent to the feedback
            receivers
            during training.
        learn_every : int, default to 1
            Time interval at which training must occur, when dealing with a
            sequence of input data. By default, the training method is called
            every time the Model receive an input.
        from_state : dict of arrays of shape ([nodes], 1, nodes.output_dim)
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Nodes states will be reset to zero before this operation.
        return_states: list of str, optional
            Names of Nodes from which to return states as output.

        Returns
        -------
        dict or array-like of shape ([n_outputs], timesteps, output_dim)
            All outputs computed during the training
            or pairs of keys and values
            where keys are output nodes names and values
            are corresponding outputs computed.
            If `call` is False,
            outputs will be null vectors.
        """

        self._check_if_only_online()

        X_, Y_ = check_xy(self, X, Y, allow_n_sequences=False)

        self._initialize_on_sequence(X_, Y_)

        states = allocate_returned_states(self, X_, return_states)

        dispatched_data = dispatch(
            X_, Y_, return_targets=True, force_teachers=force_teachers
        )

        with self.with_state(from_state, stateful=stateful, reset=reset):
            # load proxys after state update to make it have an
            # impact on feedback
            self._load_proxys(keep=True)
            for i, (x, forced_feedback, y) in enumerate(dispatched_data):

                if not force_teachers:
                    forced_feedback = None

                with self.with_feedback(forced_feedback):
                    state = self._call(x, return_states=return_states)

                # reload feedbacks from training. Some nodes may need
                # updated feedback signals to train.
                self._load_proxys()

                if i % learn_every == 0 or len(X) == 1:
                    self._train(self, x=x, y=y, force_teachers=force_teachers)

                if is_mapping(state):
                    for name, value in state.items():
                        states[name][i, :] = value
                else:
                    states[self.output_nodes[0].name][i, :] = state

        self._clean_proxys()
        self._unregister_teachers()

        # dicts are only relevant if model has several outputs (len > 1) or
        # if we want to return states from hidden nodes
        if len(states) == 1 and return_states is None:
            return states[self.output_nodes[0].name]

        return states

    def fit(
        self,
        X: MappedData,
        Y: MappedData,
        warmup=0,
        force_teachers=True,
        from_state=None,
        stateful=True,
        reset=False,
    ) -> "Model":
        """Fit all offline Nodes in the Model
        using their offline learning rule.

        Parameters
        ----------
        X : dict or array-like of shape ([n_inputs], [series], timesteps, input_dim)
            Input sequence of data. If dict, then pairs
            of keys and values, where keys are Model input
            nodes names and values are sequence of input data.
        Y : dict or array-like of shape ([offlines], [series], timesteps, offlines.output_dim)
            Target sequence of data. If dict, then pairs
            of keys and values, where keys are Model input
            nodes names and values are sequence of target data.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.
        force_teachers : bool, default to True
            If True, this Model will broadcast the available ground truth
            signal
            to all online nodes sending feedback to other nodes. Otherwise,
            the real state of these nodes will be sent to the feedback
            receivers
            during training.
        from_state : dict of arrays of shape ([nodes], 1, nodes.output_dim)
            Pairs of keys and values, where keys are Model nodes names and
            value are new ndarray state vectors.
        stateful : bool, default to True
            If True, Node state will be updated by this operation.
        reset : bool, default to False
            If True, Nodes states will be reset to zero before this operation.

        Returns
        -------
        Model
            Model trained offline.
        """

        if not any([n for n in self.trainable_nodes if n.is_trained_offline]):
            raise TypeError(
                f"Impossible to fit model {self} offline: "
                "no offline nodes found in model."
            )

        X, Y = to_data_mapping(self, X, Y)

        self._initialize_on_sequence(X[0], Y[0])
        self.initialize_buffers()

        subgraphs = get_offline_subgraphs(self.nodes, self.edges)

        trained = set()
        next_X = None

        with self.with_state(from_state, reset=reset, stateful=stateful):
            for i, ((nodes, edges), relations) in enumerate(subgraphs):
                submodel, offlines = build_forward_sumodels(nodes, edges, trained)

                if next_X is not None:
                    for j in range(len(X)):
                        X[j].update(next_X[j])

                return_states = None
                if len(relations) > 0:
                    return_states = list(relations.keys())

                # next inputs for next submodel
                next_X = []
                seq = progress(X, f"Running {self.name}")

                _partial_fit_fn = partial(
                    run_and_partial_fit,
                    force_teachers=force_teachers,
                    complete_model=self,
                    submodel=submodel,
                    warmup=warmup,
                    reset=reset,
                    offlines=offlines,
                    relations=relations,
                    stateful=stateful,
                    return_states=return_states,
                )

                # for seq/batch in dataset
                for x_seq, y_seq in zip(seq, Y):
                    next_X += [_partial_fit_fn(x_seq=x_seq, y_seq=y_seq)]

                for node in offlines:
                    if verbosity():
                        print(f"Fitting node {node.name}...")
                    node.fit()

                trained |= set(offlines)

        return self

    def copy(self, *args, **kwargs):
        raise NotImplementedError()


class FrozenModel(Model):
    """A FrozenModel is a Model that can not be
    linked to other nodes or models.
    """

    def __init__(self, *args, **kwargs):
        super(FrozenModel, self).__init__(*args, **kwargs)

    def update_graph(self, new_nodes, new_edges):
        raise TypeError(
            f"Impossible to update FrozenModel {self}: "
            f"model is frozen and cannot be modified."
        )



# Author: Nathan Trouvain at 15/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from reservoirpy.nodes import Identity, ReLU, Sigmoid, Softmax, Softplus, Tanh


@pytest.mark.parametrize(
    "node",
    (Tanh(), Softmax(), Softplus(), Sigmoid(), Identity(), ReLU(), Softmax(beta=2.0)),
)
def test_activation(node):
    x = np.ones((1, 10))
    out = node(x)
    assert out.shape == x.shape



# Author: Nathan Trouvain at 09/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np

from ..node import Node
from ..utils.validation import is_sequence_set


def forward(add, *data):
    if len(data) > 1:
        data = np.squeeze(data)
    else:  # nothing to add
        data = data[0]
    return np.sum(data, axis=0).reshape(1, -1)


def initialize(add, x=None):
    if x is not None:
        # x is an array, add over axis 0
        if isinstance(x, np.ndarray):
            add.set_input_dim(x.shape)
            add.set_output_dim((1, x.shape[1]))

        elif is_sequence_set(x):
            shapes = [array.shape for array in x]

            if not all([s[0] == 1 for s in shapes]):
                raise ValueError(
                    f"Each timestep of data must be represented "
                    f"by a vector of shape (1, dimension) when "
                    f"entering node {add.name}. Received inputs "
                    f"of shape {shapes}."
                )

            add.set_input_dim((len(x), x[0].shape[1]))

            if len(set([s[1] for s in shapes])) > 1:
                raise ValueError(
                    f"Impossible to sum inputs: inputs have "
                    f"different dimensions  entering node "
                    f"{add.name}. Received inputs of shape "
                    f"{shapes}."
                )
            else:
                add.set_output_dim((1, x[0].shape[1]))


class Add(Node):
    def __init__(self, name=None):
        super(Add, self).__init__(forward=forward, initializer=initialize, name=name)



from math import cos, sin

import numpy as np
import pytest

from reservoirpy.datasets import lorenz

from .._esn_online import ESNOnline


@pytest.fixture(scope="session")
def matrices():
    Win = np.array([[1, -1], [-1, 1], [1, -1], [-1, -1]])
    W = np.array(
        [
            [0.0, 0.1, -0.1, 0.0],
            [0.2, 0.0, 0.0, -0.2],
            [0.0, 0.2, 0.3, 0.1],
            [-0.1, 0.0, 0.0, 0.0],
        ]
    )
    Wout = np.zeros((2, 4 + 1))

    return W, Win, Wout


@pytest.fixture(scope="session")
def matrices_fb():
    Win = np.array([[1, -1], [-1, 1], [1, -1], [-1, -1]])
    W = np.array(
        [
            [0.0, 0.1, -0.1, 0.0],
            [0.2, 0.0, 0.0, -0.2],
            [0.0, 0.2, 0.3, 0.1],
            [-0.1, 0.0, 0.0, 0.0],
        ]
    )
    Wfb = np.array([[1, -1], [-1, -1], [1, 1], [-1, 1]])
    Wout = np.zeros((2, 4 + 1))
    return W, Win, Wout, Wfb


@pytest.fixture(scope="session")
def dummy_data():
    Xn0 = np.array([[sin(x), cos(x)] for x in np.linspace(0, 4 * np.pi, 500)])
    Xn1 = np.array(
        [[sin(x), cos(x)] for x in np.linspace(np.pi / 4, 4 * np.pi + np.pi / 4, 500)]
    )
    return Xn0, Xn1


def test_esn(matrices, dummy_data):
    W, Win, Wout = matrices
    esn = ESNOnline(lr=0.1, W=W, Win=Win, dim_out=2, input_bias=False)
    X, y = dummy_data
    states = esn.train([X], [y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X])

    assert states[0].shape[0] == X.shape[0]
    assert outputs[0].shape[1] == y.shape[1]

    states = esn.train([X, X, X], [y, y, y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X, X])

    assert len(states) == 2
    assert len(outputs) == 2


def test_esn_fb(matrices_fb, dummy_data):
    W, Win, Wout, Wfb = matrices_fb
    esn = ESNOnline(
        lr=0.1, W=W, Win=Win, Wfb=Wfb, dim_out=2, input_bias=False, fbfunc=np.tanh
    )
    X, y = dummy_data
    states = esn.train([X], [y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X])

    assert states[0].shape[0] == X.shape[0]
    assert outputs[0].shape[1] == y.shape[1]

    states = esn.train([X, X, X], [y, y, y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X, X])

    assert len(states) == 2
    assert len(outputs) == 2



# Author: Nathan Trouvain at 06/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial

from ..activationsfunc import get_function
from ..node import Node


def forward(node: Node, x, **kwargs):
    return node.f(x, **kwargs)


def initialize(node: Node, x=None, *args, **kwargs):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


class Softmax(Node):
    """Softmax activation function.

    .. math::

        y_k = \\frac{e^{\\beta x_k}}{\\sum_{i=1}^{n} e^{\\beta x_i}}

    :py:attr:`Softmax.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.softmax`).
    ``beta``      Softmax :math:`\\beta` parameter (1.0 by default).
    ============= ======================================================================

    Parameters
    ----------
    beta: float, default to 1.0
        Beta parameter of softmax.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, beta=1.0, **kwargs):
        super(Softmax, self).__init__(
            hypers={"f": get_function("softmax"), "beta": beta},
            forward=partial(forward, beta=beta),
            initializer=initialize,
            **kwargs,
        )


class Softplus(Node):
    """Softplus activation function.

    .. math::

        f(x) = \\mathrm{ln}(1 + e^{x})

    :py:attr:`Softplus.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.softplus`).
    ============= ======================================================================

    Parameters
    ----------
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, **kwargs):
        super(Softplus, self).__init__(
            hypers={"f": get_function("softplus")},
            forward=forward,
            initializer=initialize,
            **kwargs,
        )


class Sigmoid(Node):
    """Sigmoid activation function.

    .. math::

        f(x) = \\frac{1}{1 + e^{-x}}

    :py:attr:`Sigmoid.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.sigmoid`).
    ============= ======================================================================

    Parameters
    ----------
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, **kwargs):
        super(Sigmoid, self).__init__(
            hypers={"f": get_function("sigmoid")},
            forward=forward,
            initializer=initialize,
            **kwargs,
        )


class Tanh(Node):
    """Hyperbolic tangent activation function.

    .. math::

        f(x) = \\frac{e^x - e^{-x}}{e^x + e^{-x}}

    :py:attr:`Tanh.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.tanh`).
    ============= ======================================================================

    Parameters
    ----------
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, **kwargs):
        super(Tanh, self).__init__(
            hypers={"f": get_function("tanh")},
            forward=forward,
            initializer=initialize,
            **kwargs,
        )


class Identity(Node):
    """Identity function.

    .. math::

        f(x) = x

    Provided for convenience.

    :py:attr:`Identity.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.identity`).
    ============= ======================================================================

    Parameters
    ----------
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, **kwargs):
        super(Identity, self).__init__(
            hypers={"f": get_function("identity")},
            forward=forward,
            initializer=initialize,
            **kwargs,
        )


class ReLU(Node):
    """ReLU activation function.

    .. math::

        f(x) = x ~~ \\mathrm{if} ~~ x > 0 ~~ \\mathrm{else} ~~ 0

    :py:attr:`ReLU.hypers` **list**

    ============= ======================================================================
    ``f``         Activation function (:py:func:`reservoir.activationsfunc.relu`).
    ============= ======================================================================

    Parameters
    ----------
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    """

    def __init__(self, **kwargs):
        super(ReLU, self).__init__(
            hypers={"f": get_function("relu")},
            forward=forward,
            initializer=initialize,
            **kwargs,
        )



"""
==================================================
ReservoirPy v0.2 tools (:mod:`reservoirpy.compat`)
==================================================

ReservoirPy v0.2 tools, kept for compatibility with older projects.

Warning
-------

    ReservoirPy v0.2 tools are deprecated. No removal date has been set,
    however, we encourage users to switch to newer versions (v0.3) of the
    library. The function :py:func:`load_compat` can help you load v0.2 models
    using the new Node API (see :ref:`node`), introduced in version v0.3.

Load and backport
=================

.. autosummary::
    :toctree: generated/

    load_compat - Load v0.2 models into v0.3
    load - Load old v0.2 models

ESN and ESNOnline
=================

.. autosummary::
    :toctree: generated/

    ESN - ESN model with parallelized training
    ESNOnline - ESN with FORCE learning

Regression models
=================

.. autosummary::
    :toctree: generated/

    regression_models.RidgeRegression - Tikhonov regression solver
"""
import json
import pathlib
import re
from typing import Union

import dill
import numpy as np
from scipy import sparse

from ..activationsfunc import identity
from ..mat_gen import zeros
from ..nodes import ESN as ESN_v3
from ..nodes import Reservoir, Ridge
from . import regression_models
from ._esn import ESN
from ._esn_online import ESNOnline
from .utils.save import load


def _load_files_from_v2(dirpath):
    dirpath = pathlib.Path(dirpath)
    matrices = dict()
    config = dict()
    for filename in dirpath.iterdir():
        ext = filename.suffix
        # all rpy <= 0.2.4 versions have file names starting with esn
        is_esn_obj = "esn" in filename.name[:3]
        if is_esn_obj:
            if ext in (".npy", ".npz"):
                matrix_name = ("Win", "W", "Wfb", "Wout", "_W", "_Win", "_Wfb", "_Wout")
                match = re.findall("_?W.*?(?=-)", filename.name)
                for name in matrix_name:
                    if name in match:
                        n = name
                        if name.startswith("_"):
                            n = name[1:]
                        matrices[n] = _load_matrix_v2(filename)
            elif ext == ".json":
                with filename.open(mode="r+") as fp:
                    config = json.load(fp)

    fns = dict()
    for attr, value in config.items():
        if attr == "fbfunc" and value is not None:
            filename = pathlib.Path(value)
            if filename.exists():
                fns["fbfunc"] = dill.load(filename)

    return matrices, fns, config


def _load_matrix_v2(filename):
    ext = filename.suffix
    mat = None
    if ext == ".npy":
        mat = np.load(str(filename))
    elif ext == ".npz":  # maybe a scipy sparse array
        try:
            mat = sparse.load_npz(str(filename))
        except Exception as e:
            mat = np.load(str(filename))
            keys = list(mat.keys())
            sparse_keys = ("indices", "indptr", "format", "shape", "data")
            if any([k in sparse_keys for k in keys]):
                raise e
            elif len(keys) == 1:  # Only one array per file
                mat = mat[keys[0]]
            else:
                raise TypeError("Unknown array format in file {filename}.")
    return mat


def load_compat(directory: Union[str, pathlib.Path]) -> ESN_v3:
    """Load a ReservoirPy v0.2.4 and lower ESN model as a
    ReservoirPy v0.3 model.

    .. warning::
        Models and Nodes should now
        be saved using Python serialization utilities
        `pickle`.

    Parameters
    ----------
    directory : str or Path

    Returns
    -------
    reservoirpy.nodes.ESN
        A ReservoirPy v0.3 ESN instance.
    """
    dirpath = pathlib.Path(directory)
    if not dirpath.exists():
        raise NotADirectoryError(f"'{directory}' not found.")

    matrices, fns, config = _load_files_from_v2(dirpath)

    attr = config.get("attr", config)

    version = config.get("version")

    msg = "Impossible to load ESN from version {} of reservoirpy: unknown model {}"
    ridge = 0.0
    if attr.get("sklearn_model") is not None:
        raise TypeError(msg.format(version, attr["sklearn_model"]))
    elif attr.get("_ridge") is not None:
        ridge = attr["_ridge"]

    if attr.get("reg_model") is not None:
        reg_model = attr["reg_model"]
        if reg_model["type"] not in ("ridge", "pinv"):
            raise TypeError(msg.format(version, attr["type"]))
        elif reg_model["type"] == "ridge":
            ridge = reg_model.get("coef", 0.0)

    feedback = False
    if matrices.get("Wfb") is not None:
        feedback = True

    output_dim = attr.get("dim_out", attr.get("_dim_out"))

    reservoir = Reservoir(
        units=attr.get("N", attr.get("_N")),
        lr=attr["lr"],
        input_bias=attr.get("in_bias", attr.get("_input_bias")),
        W=matrices["W"],
        Win=matrices["Win"],
        Wfb=matrices.get("Wfb"),
        fb_activation=fns.get("fbfunc", identity),
        noise_in=attr.get("noise_in", 0.0),
        noise_rc=attr.get("noise_rc", 0.0),
        noise_fb=attr.get("noise_out", 0.0),
        noise_type="uniform",
        seed=attr.get("seed"),
    )

    W = matrices.get("Wout")
    if W is None:
        Wout = zeros
        bias = zeros
    else:
        Wout = W[:, 1:]
        bias = W[:, :1]

    readout = Ridge(
        output_dim=output_dim, ridge=ridge, Wout=Wout, bias=bias, input_bias=True
    )

    model = ESN_v3(reservoir=reservoir, readout=readout, feedback=feedback)

    return model


__all__ = ["ESN", "ESNOnline", "load_compat", "regression_models", "load"]



# Author: Nathan Trouvain at 06/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial

import numpy as np

from ..node import Node
from ..nodes.readouts.base import (
    _initialize_readout,
    _prepare_inputs_for_learning,
    readout_forward,
)


def _reset_buffers(step, rTPs, factors):
    step[:] = np.zeros_like(step)
    rTPs[:] = np.zeros_like(rTPs)
    factors[:] = np.zeros_like(factors)


def train(readout: "BatchFORCE", x=None, y=None):

    if x is not None:

        x, y = _prepare_inputs_for_learning(
            x, y, bias=readout.has_bias, allow_reshape=True
        )

        W = readout.Wout
        if readout.has_bias:
            bias = readout.bias
            W = np.c_[bias, W]

        P = readout.P
        r = x.T
        output = readout.state()

        factors = readout.get_buffer("factors")
        rTPs = readout.get_buffer("rTPs")
        steps = readout.get_buffer("step")
        step = int(steps[0])

        error = output.T - y.T

        rt = r.T
        rTP = (rt @ P) - (rt @ (factors * rTPs)) @ rTPs.T
        factor = float(1.0 / (1.0 + rTP @ r))

        factors[step] = factor
        rTPs[:, step] = rTP

        new_rTP = rTP * (1 - factor * (rTP @ r).item())

        W -= error @ new_rTP

        if readout.has_bias:
            readout.set_param("Wout", W[:, 1:])
            readout.set_param("bias", W[:, :1])
        else:
            readout.set_param("Wout", W)

        step += 1

        if step == readout.batch_size:
            P -= (factors * rTPs) @ rTPs.T
            _reset_buffers(steps, rTPs, factors)


def initialize(readout: "BatchFORCE", x=None, y=None, init_func=None, bias=None):

    _initialize_readout(readout, x, y, init_func, bias)

    if x is not None:
        input_dim, alpha = readout.input_dim, readout.alpha

        if readout.has_bias:
            input_dim += 1

        P = np.asmatrix(np.eye(input_dim)) / alpha

        readout.set_param("P", P)


def initialize_buffers(readout: "BatchFORCE"):
    bias_dim = 0
    if readout.has_bias:
        bias_dim = 1

    readout.create_buffer("rTPs", (readout.input_dim + bias_dim, readout.batch_size))
    readout.create_buffer("factors", (readout.batch_size,))
    readout.create_buffer("step", (1,))


class BatchFORCE(Node):

    # A special thanks to Lionel Eyraud-Dubois and
    # Olivier Beaumont for their improvement of this method.

    def __init__(
        self,
        output_dim=None,
        alpha=1e-6,
        batch_size=1,
        Wout_init=np.zeros,
        bias=True,
        name=None,
    ):
        super(BatchFORCE, self).__init__(
            params={"Wout": None, "bias": None, "P": None},
            hypers={"alpha": alpha, "batch_size": batch_size, "has_bias": bias},
            forward=readout_forward,
            train=train,
            initializer=partial(initialize, init_func=Wout_init, bias=bias),
            buffers_initializer=initialize_buffers,
            output_dim=output_dim,
            name=name,
        )



# Author: Nathan Trouvain at 14/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Any, Sequence, Union

import numpy as np
from scipy.sparse import issparse

from reservoirpy.utils.validation import check_vector


def _check_values(array_or_list: Union[Sequence, np.ndarray], value: Any):
    """Check if the given array or list contains the given value."""
    if value == np.nan:
        assert (
            np.isnan(array_or_list).any() == False
        ), f"{array_or_list} should not contain NaN values."
    if value is None:
        if type(array_or_list) is list:
            assert (
                np.count_nonzero(array_or_list == None) == 0
            ), f"{array_or_list} should not contain None values."
        elif type(array_or_list) is np.array:
            # None is transformed to np.nan when it is in an array
            assert (
                np.isnan(array_or_list).any() == False
            ), f"{array_or_list} should not contain NaN values."


def check_input_lists(X, dim_in, Y=None, dim_out=None):
    if isinstance(X, np.ndarray):
        X = [X]

    if Y is not None:
        if isinstance(Y, np.ndarray):
            Y = [Y]
        if not (len(X) == len(Y)):
            raise ValueError(
                f"Inconsistent number of inputs and targets: "
                f"found {len(X)} input sequences, but {len(Y)} "
                f"target sequences."
            )

    for i in range(len(X)):
        x = check_vector(X[i], allow_reshape=False)

        if x.ndim != 2:
            raise ValueError(
                f"Input {i} has shape {x.shape} but should "
                f"be 2-dimensional, with first axis representing "
                f"time and second axis representing features."
            )

        if x.shape[1] != dim_in:
            raise ValueError(
                f"Input {i} has {x.shape[1]} features but ESN expects "
                f"{dim_in} features as input."
            )

        if Y is not None:
            y = check_vector(Y[i], allow_reshape=False)
            if y.ndim != 2:
                raise ValueError(
                    f"Target {i} has shape {y.shape} but should "
                    f"be 2-dimensional, with first axis "
                    f"representing "
                    f"time and second axis representing "
                    f"features."
                )

            if x.shape[0] != y.shape[0]:
                raise ValueError(
                    f"Inconsistent inputs and targets lengths: "
                    f"input {i} has length {x.shape[0]} but "
                    f"corresponding target {i} has length "
                    f"{y.shape[0]}."
                )

            if dim_out is not None:
                if y.shape[1] != dim_out:
                    raise ValueError(
                        f"Target {i} has {y.shape[1]} features but ESN "
                        f"expects "
                        f"{dim_out} features as feedback."
                    )

    return X, Y


def check_reservoir_matrices(W, Win, Wout=None, Wfb=None, caller=None):
    caller_name = f"{caller.__class__.__name__} :" if caller is not None else ""

    W = check_datatype(W, caller=caller, name="W")
    Win = check_datatype(Win, caller=caller, name="Win")

    in_shape = Win.shape
    res_shape = W.shape

    # W shape is (units, units)
    if res_shape[0] != res_shape[1]:
        raise ValueError(
            f"{caller_name} reservoir matrix W should be square but has "
            f"shape {res_shape}."
        )
    # Win shape is (units, dim_in [+ bias])
    if in_shape[0] != res_shape[0]:
        raise ValueError(
            f"{caller_name} dimension mismatch between W and Win: "
            f"W is of shape {res_shape} and Win is of shape {in_shape} "
            f"({res_shape[0]} != {in_shape[0]})."
        )

    # Wout shape is (dim_out, units + bias)
    out_shape = None
    if Wout is not None:
        Wout = check_datatype(Wout, caller=caller, name="Wout")
        out_shape = Wout.shape
        if out_shape[1] != res_shape[0] + 1:
            raise ValueError(
                f"{caller_name} dimension mismatch between W and Wout: "
                f"W is of shape {res_shape} and Wout is of shape {out_shape} "
                f"({res_shape[0]} + bias (1) != {out_shape[1]})."
            )
    # Wfb shape is (units, dim_out)
    if Wfb is not None:
        Wfb = check_datatype(Wfb, caller=caller, name="Wfb")
        fb_shape = Wfb.shape
        if out_shape is not None:
            if fb_shape[1] != out_shape[0]:
                raise ValueError(
                    f"{caller_name} dimension mismatch between Wfb and Wout: "
                    f"Wfb is of shape {fb_shape} and Wout is of sh"
                    f"ape {out_shape} "
                    f"({fb_shape[1]} != {out_shape[0]})."
                )
        if fb_shape[0] != res_shape[0]:
            raise ValueError(
                f"{caller_name} dimension mismatch between W and Wfb: "
                f"W is of shape {res_shape} and Wfb is of shape {fb_shape} "
                f"({res_shape[0]} != {fb_shape[0]})."
            )

    return W, Win, Wout, Wfb


def check_datatype(array, caller=None, name=None, allow_inf=False, allow_nan=False):
    caller_name = f"{caller.__class__.__name__} :" if caller is not None else ""
    array_name = name if isinstance(name, str) else array.__class__.__name___

    if not isinstance(array, np.ndarray) and not issparse(array):
        array = np.asarray(array)

    if not (np.issubdtype(array.dtype, np.number)):
        raise TypeError(
            f"{caller_name} Impossible to operate on non-numerical data, "
            f"in array '{array_name}' of type {array.dtype}: {array}"
        )

    if not allow_nan:
        msg = (
            f"{caller_name} Impossible to operate on NaN value, "
            f"in array '{array_name}': {array}."
        )
        if issparse(array):
            if np.any(np.isnan(array.data)):
                raise ValueError(msg)
        else:
            if np.any(np.isnan(array)):
                raise ValueError(msg)

    if not allow_inf:
        msg = (
            f"{caller_name} Impossible to operate on inf value, "
            f"in array '{array_name}': {array}."
        )
        if issparse(array):
            if np.any(np.isinf(array.data)):
                raise ValueError(msg)
        else:
            if np.any(np.isinf(array)):
                raise ValueError(msg)

    return array



# Author: Nathan Trouvain at 16/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from functools import partial
from typing import Callable, Dict, Optional, Sequence, Union

from ...activationsfunc import get_function, identity, tanh
from ...mat_gen import bernoulli, normal
from ...node import Node
from ...type import Weights
from ...utils.random import noise, rand_generator
from ...utils.validation import is_array
from .base import forward_external, forward_internal, initialize, initialize_feedback


class Reservoir(Node):
    """Pool of leaky-integrator neurons with random recurrent connexions.

    Reservoir neurons states, gathered in a vector :math:`\\mathbf{x}`, may follow
    one of the two update rules below:

    - **1.** Activation function is part of the neuron internal state
      (equation called ``internal``):

    .. math::

        \\mathbf{x}[t+1] = (1 - \\mathrm{lr}) * \\mathbf{x}[t] + \\mathrm{lr}
         * f(\\mathbf{W}_{in} \\cdot (\\mathbf{u}[t+1]+c_{in}*\\xi)
          + \\mathbf{W} \\cdot \\mathbf{x}[t]
        + \\mathbf{W}_{fb} \\cdot (g(\\mathbf{y}[t])+c_{fb}*\\xi) + \\mathbf{b})
        + c * \\xi

    - **2.** Activation function is applied on emitted internal states
      (equation called ``external``):

    .. math::


        \\mathbf{r}[t+1] = (1 - \\mathrm{lr}) * \\mathbf{r}[t] + \\mathrm{lr}
        * (\\mathbf{W}_{in} \\cdot (\\mathbf{u}[t+1]+c_{in}*\\xi)
         + \\mathbf{W} \\cdot \\mathbf{x}[t]
        + \\mathbf{W}_{fb} \\cdot (g(\\mathbf{y}[t])+c_{fb}*\\xi) + \\mathbf{b})

    .. math::

        \\mathbf{x}[t+1] = f(\\mathbf{r}[t+1]) + c * \\xi

    where:
        - :math:`\\mathbf{x}` is the output activation vector of the reservoir;
        - :math:`\\mathbf{r}` is the (optional) internal activation vector of the reservoir;
        - :math:`\\mathbf{u}` is the input timeseries;
        - :math:`\\mathbf{y}` is a feedback vector;
        - :math:`\\xi` is a random noise;
        - :math:`f` and :math:`g` are activation functions.

    :py:attr:`Reservoir.params` **list:**

    ================== ===================================================================
    ``W``              Recurrent weights matrix (:math:`\\mathbf{W}`).
    ``Win``            Input weights matrix (:math:`\\mathbf{W}_{in}`).
    ``Wfb``            Feedback weights matrix (:math:`\\mathbf{W}_{fb}`).
    ``bias``           Input bias vector (:math:`\\mathbf{b}`).
    ``internal_state``  Internal state used with equation="external" (:math:`\\mathbf{r}`).
    ================== ===================================================================

    :py:attr:`Reservoir.hypers` **list:**

    ======================= ========================================================
    ``lr``                  Leaking rate (1.0 by default) (:math:`\\mathrm{lr}`).
    ``sr``                  Spectral radius of ``W`` (optional).
    ``input_scaling``       Input scaling (float or array) (1.0 by default).
    ``fb_scaling``          Feedback scaling (float or array) (1.0 by default).
    ``rc_connectivity``     Connectivity (or density) of ``W`` (0.1 by default).
    ``input_connectivity``  Connectivity (or density) of ``Win`` (0.1 by default).
    ``fb_connectivity``     Connectivity (or density) of ``Wfb`` (0.1 by default).
    ``noise_in``            Input noise gain (0 by default) (:math:`c_{in} * \\xi`).
    ``noise_rc``            Reservoir state noise gain (0 by default) (:math:`c * \\xi`).
    ``noise_fb``            Feedback noise gain (0 by default) (:math:`c_{fb} * \\xi`).
    ``noise_type``          Distribution of noise (normal by default) (:math:`\\xi \\sim \\mathrm{Noise~type}`).
    ``activation``          Activation of the reservoir units (tanh by default) (:math:`f`).
    ``fb_activation``       Activation of the feedback units (identity by default) (:math:`g`).
    ``units``               Number of neuronal units in the reservoir.
    ``noise_generator``     A random state generator.
    ======================= ========================================================

    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    sr : float, optional
        Spectral radius of recurrent weight matrix.
    input_bias : bool, default to True
        If False, no bias is added to inputs.
    noise_rc : float, default to 0.0
        Gain of noise applied to reservoir activations.
    noise_in : float, default to 0.0
        Gain of noise applied to input inputs.
    noise_fb : float, default to 0.0
        Gain of noise applied to feedback signal.
    noise_type : str, default to "normal"
        Distribution of noise. Must be a Numpy random variable generator
        distribution (see :py:class:`numpy.random.Generator`).
    noise_kwargs : dict, optional
        Keyword arguments to pass to the noise generator, such as `low` and `high`
        values of uniform distribution.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    bias_scaling: float, default to 1.0
        Bias gain.
    fb_scaling : float or array-like of shape (features,), default to 1.0
        Feedback gain. An array of the same dimension as the feedback can be used to
        set up different feedback scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    fb_connectivity : float, default to 0.1
        Connectivity of feedback neurons, i.e. ratio of feedback neurons
        connected to reservoir neurons. Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.normal`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    Wfb : callable or array-like of shape (units, feedback), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Feedback weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    fb_activation : str or callable, default to :py:func:`~reservoirpy.activationsfunc.identity`
        Feedback activation function.
        - If a str, should be a :py:mod:`~reservoirpy.activationsfunc`
        function name.
        - If a callable, should be an element-wise operator on arrays.
    activation : str or callable, default to :py:func:`~reservoirpy.activationsfunc.tanh`
        Reservoir units activation function.
        - If a str, should be a :py:mod:`~reservoirpy.activationsfunc`
        function name.
        - If a callable, should be an element-wise operator on arrays.
    equation : {"internal", "external"}, default to "internal"
        If "internal", will use equation defined in equation 1 to update the state of
        reservoir units. If "external", will use the equation defined in equation 2
        (see above).
    feedback_dim : int, optional
        Feedback dimension. Can be inferred at first call.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.

    Note
    ----

    If W, Win, bias or Wfb are initialized with an array-like matrix, then all
    initializers parameters such as spectral radius (``sr``) or input scaling
    (``input_scaling``) are ignored.
    See :py:mod:`~reservoirpy.mat_gen` for more information.

    Example
    -------

    >>> from reservoirpy.nodes import Reservoir
    >>> reservoir = Reservoir(100, lr=0.2, sr=0.8) # a 100 neurons reservoir

    Using the :py:func:`~reservoirpy.datasets.mackey_glass` timeseries:

    >>> from reservoirpy.datasets import mackey_glass
    >>> x = mackey_glass(200)
    >>> states = reservoir.run(x)

    .. plot::

        from reservoirpy.nodes import Reservoir
        reservoir = Reservoir(100, lr=0.2, sr=0.8)
        from reservoirpy.datasets import mackey_glass
        x = mackey_glass(200)
        states = reservoir.run(x)
        fig, ax = plt.subplots(6, 1, figsize=(7, 10), sharex=True)
        ax[0].plot(x)
        ax[0].grid()
        ax[0].set_title("Neuron states (on Mackey-Glass)")
        for i in range(1, 6):
            ax[i].plot(states[:, i], label=f"Neuron {i}")
            ax[i].legend()
            ax[i].grid()
        ax[-1].set_xlabel("Timesteps")
    """

    def __init__(
        self,
        units: int = None,
        lr: float = 1.0,
        sr: Optional[float] = None,
        input_bias: bool = True,
        noise_rc: float = 0.0,
        noise_in: float = 0.0,
        noise_fb: float = 0.0,
        noise_type: str = "normal",
        noise_kwargs: Dict = None,
        input_scaling: Union[float, Sequence] = 1.0,
        bias_scaling: float = 1.0,
        fb_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: float = 0.1,
        rc_connectivity: float = 0.1,
        fb_connectivity: float = 0.1,
        Win: Union[Weights, Callable] = bernoulli,
        W: Union[Weights, Callable] = normal,
        Wfb: Union[Weights, Callable] = bernoulli,
        bias: Union[Weights, Callable] = bernoulli,
        fb_activation: Union[str, Callable] = identity,
        activation: Union[str, Callable] = tanh,
        equation: Literal["internal", "external"] = "internal",
        input_dim: Optional[int] = None,
        feedback_dim: Optional[int] = None,
        seed=None,
        **kwargs,
    ):
        if units is None and not is_array(W):
            raise ValueError(
                "'units' parameter must not be None if 'W' parameter is not "
                "a matrix."
            )

        if equation == "internal":
            forward = forward_internal
        elif equation == "external":
            forward = forward_external
        else:
            raise ValueError(
                "'equation' parameter must be either 'internal' or 'external'."
            )

        if type(activation) is str:
            activation = get_function(activation)
        if type(fb_activation) is str:
            fb_activation = get_function(fb_activation)

        rng = rand_generator(seed)

        noise_kwargs = dict() if noise_kwargs is None else noise_kwargs

        super(Reservoir, self).__init__(
            fb_initializer=partial(
                initialize_feedback,
                Wfb_init=Wfb,
                fb_scaling=fb_scaling,
                fb_connectivity=fb_connectivity,
                seed=seed,
            ),
            params={
                "W": None,
                "Win": None,
                "Wfb": None,
                "bias": None,
                "internal_state": None,
            },
            hypers={
                "lr": lr,
                "sr": sr,
                "input_scaling": input_scaling,
                "bias_scaling": bias_scaling,
                "fb_scaling": fb_scaling,
                "rc_connectivity": rc_connectivity,
                "input_connectivity": input_connectivity,
                "fb_connectivity": fb_connectivity,
                "noise_in": noise_in,
                "noise_rc": noise_rc,
                "noise_out": noise_fb,
                "noise_type": noise_type,
                "activation": activation,
                "fb_activation": fb_activation,
                "units": units,
                "noise_generator": partial(noise, rng=rng, **noise_kwargs),
            },
            forward=forward,
            initializer=partial(
                initialize,
                sr=sr,
                input_scaling=input_scaling,
                bias_scaling=bias_scaling,
                input_connectivity=input_connectivity,
                rc_connectivity=rc_connectivity,
                W_init=W,
                Win_init=Win,
                bias_init=bias,
                input_bias=input_bias,
                seed=seed,
            ),
            output_dim=units,
            feedback_dim=feedback_dim,
            input_dim=input_dim,
            **kwargs,
        )



# Author: Paul Bernard on 10/01/2024 <paul.bernard@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import deque
from functools import partial

import numpy as np

from ..node import Node


def forward(node: Node, x, **kwargs):
    node.buffer.appendleft(x)
    output = node.buffer.pop()

    return output


def initialize(node: Node, x=None, y=None, initial_values=None, *args, **kwargs):
    if node.input_dim is not None:
        dim = node.input_dim
    else:
        dim = x.shape[1]

    node.set_input_dim(dim)
    node.set_output_dim(dim)

    if initial_values is None:
        initial_values = np.zeros((node.delay, node.input_dim), dtype=node.dtype)
    node.set_param("buffer", deque(initial_values, maxlen=node.delay + 1))


class Delay(Node):
    """Delays the data transmitted through this node without transformation.

    :py:attr:`Delay.params` **list**

    ============= ======================================================================
    ``buffer``    (:py:class:`collections.deque`) Buffer of the values coming next.
    ============= ======================================================================

    :py:attr:`Delay.hypers` **list**

    ============= ======================================================================
    ``delay``     (int) Number of timesteps before outputting the input.
    ============= ======================================================================

    Parameters
    ----------
    delay: int, defaults to 1.
        Number of timesteps before outputting the input.
    initial_values: array of shape (delay, input_dim), defaults to
        `np.zeros((delay, input_dim))`.
        Initial outputs of the node.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    dtype : Numpy dtype, defaults to np.float64
        Numerical type for node parameters.
    name : str, optional
        Node name.

    Examples
    --------
    >>> x = np.arange(10.).reshape(-1, 1)
    >>> from reservoirpy.nodes import Delay
    >>> delay_node = Delay(
    ...     delay=3,
    ...     initial_values=np.array([[-3.0], [-2.0], [-1.0]])
    ... )
    >>> out = delay_node.run(x)
    >>> print(out.T)
    [[-1. -2. -3.  0.  1.  2.  3.  4.  5.  6.]]
    >>> print(delay_node.buffer)
    deque([array([[9.]]), array([[8.]]), array([[7.]])], maxlen=4)
    """

    def __init__(
        self,
        delay=1,
        initial_values=None,
        input_dim=None,
        dtype=None,
        **kwargs,
    ):
        if input_dim is None and type(initial_values) is np.ndarray:
            input_dim = initial_values.shape[-1]

        super(Delay, self).__init__(
            hypers={"delay": delay},
            params={"buffer": None},
            forward=forward,
            initializer=partial(initialize, initial_values=initial_values),
            input_dim=input_dim,
            **kwargs,
        )



# Author: Nathan Trouvain at 08/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np

from ...mat_gen import zeros
from ...utils.validation import is_array


def reservoir_kernel(reservoir, u, r):
    """Reservoir base forward function.

    Computes: s[t+1] = W.r[t] + Win.(u[t] + noise) + Wfb.(y[t] + noise) + bias
    """
    W = reservoir.W
    Win = reservoir.Win
    bias = reservoir.bias

    g_in = reservoir.noise_in
    dist = reservoir.noise_type
    noise_gen = reservoir.noise_generator

    pre_s = W @ r + Win @ (u + noise_gen(dist=dist, shape=u.shape, gain=g_in)) + bias

    if reservoir.has_feedback:
        Wfb = reservoir.Wfb
        g_fb = reservoir.noise_out
        h = reservoir.fb_activation

        y = reservoir.feedback().reshape(-1, 1)
        y = h(y) + noise_gen(dist=dist, shape=y.shape, gain=g_fb)

        pre_s += Wfb @ y

    return np.array(pre_s)


def forward_internal(reservoir, x: np.ndarray) -> np.ndarray:
    """Reservoir with internal activation function:

    .. math::

        r[n+1] = (1 - \\alpha) \\cdot r[t] + \\alpha
         \\cdot f (W_{in} \\cdot u[n] + W \\cdot r[t])


    where :math:`r[n]` is the state and the output of the reservoir."""
    lr = reservoir.lr
    f = reservoir.activation
    dist = reservoir.noise_type
    g_rc = reservoir.noise_rc
    noise_gen = reservoir.noise_generator

    u = x.reshape(-1, 1)
    r = reservoir.state().T

    s_next = (
        np.multiply((1 - lr), r.T).T
        + np.multiply(lr, f(reservoir_kernel(reservoir, u, r)).T).T
        + noise_gen(dist=dist, shape=r.shape, gain=g_rc)
    )

    return s_next.T


def forward_external(reservoir, x: np.ndarray) -> np.ndarray:
    """Reservoir with external activation function:

    .. math::

        x[n+1] = (1 - \\alpha) \\cdot x[t] + \\alpha
         \\cdot f (W_{in} \\cdot u[n] + W \\cdot r[t])

        r[n+1] = f(x[n+1])


    where :math:`x[n]` is the internal state of the reservoir and :math:`r[n]`
    is the response of the reservoir."""
    lr = reservoir.lr
    f = reservoir.activation
    dist = reservoir.noise_type
    g_rc = reservoir.noise_rc
    noise_gen = reservoir.noise_generator

    u = x.reshape(-1, 1)
    r = reservoir.state().T
    s = reservoir.internal_state.T

    s_next = (
        np.multiply((1 - lr), s.T).T
        + np.multiply(lr, reservoir_kernel(reservoir, u, r).T).T
        + noise_gen(dist=dist, shape=r.shape, gain=g_rc)
    )

    reservoir.set_param("internal_state", s_next.T)

    return f(s_next).T


def initialize(
    reservoir,
    x=None,
    y=None,
    sr=None,
    input_scaling=None,
    bias_scaling=None,
    input_connectivity=None,
    rc_connectivity=None,
    W_init=None,
    Win_init=None,
    bias_init=None,
    input_bias=None,
    seed=None,
):
    if x is not None:
        reservoir.set_input_dim(x.shape[1])

        dtype = reservoir.dtype
        dtype_msg = (
            "Data type {} not understood in {}. {} should be an array or a "
            "callable returning an array."
        )

        if is_array(W_init):
            W = W_init
            if W.shape[0] != W.shape[1]:
                raise ValueError(
                    "Dimension mismatch inside W: "
                    f"W is {W.shape} but should be "
                    f"a square matrix."
                )

            if W.shape[0] != reservoir.output_dim:
                reservoir._output_dim = W.shape[0]
                reservoir.hypers["units"] = W.shape[0]

        elif callable(W_init):
            W = W_init(
                reservoir.output_dim,
                reservoir.output_dim,
                sr=sr,
                connectivity=rc_connectivity,
                dtype=dtype,
                seed=seed,
            )
        else:
            raise ValueError(dtype_msg.format(str(type(W_init)), reservoir.name, "W"))

        reservoir.set_param("units", W.shape[0])
        reservoir.set_param("W", W.astype(dtype))

        out_dim = reservoir.output_dim

        Win_has_bias = False
        if is_array(Win_init):
            Win = Win_init

            msg = (
                f"Dimension mismatch in {reservoir.name}: Win input dimension is "
                f"{Win.shape[1]} but input dimension is {x.shape[1]}."
            )

            # is bias vector inside Win ?
            if Win.shape[1] == x.shape[1] + 1:
                if input_bias:
                    Win_has_bias = True
                else:
                    bias_msg = (
                        " It seems Win has a bias column, but 'input_bias' is False."
                    )
                    raise ValueError(msg + bias_msg)
            elif Win.shape[1] != x.shape[1]:
                raise ValueError(msg)

            if Win.shape[0] != out_dim:
                raise ValueError(
                    f"Dimension mismatch in {reservoir.name}: Win internal dimension "
                    f"is {Win.shape[0]} but reservoir dimension is {out_dim}"
                )

        elif callable(Win_init):
            Win = Win_init(
                reservoir.output_dim,
                x.shape[1],
                input_scaling=input_scaling,
                connectivity=input_connectivity,
                dtype=dtype,
                seed=seed,
            )
        else:
            raise ValueError(
                dtype_msg.format(str(type(Win_init)), reservoir.name, "Win")
            )

        if input_bias:
            if not Win_has_bias:
                if callable(bias_init):
                    bias = bias_init(
                        reservoir.output_dim,
                        1,
                        input_scaling=bias_scaling,
                        connectivity=input_connectivity,
                        dtype=dtype,
                        seed=seed,
                    )
                elif is_array(bias_init):
                    bias = bias_init
                    if bias.shape[0] != reservoir.output_dim or (
                        bias.ndim > 1 and bias.shape[1] != 1
                    ):
                        raise ValueError(
                            f"Dimension mismatch in {reservoir.name}: bias shape is "
                            f"{bias.shape} but should be {(reservoir.output_dim, 1)}"
                        )
                else:
                    raise ValueError(
                        dtype_msg.format(str(type(bias_init)), reservoir.name, "bias")
                    )
            else:
                bias = Win[:, :1]
                Win = Win[:, 1:]
        else:
            bias = zeros(reservoir.output_dim, 1, dtype=dtype)

        reservoir.set_param("Win", Win.astype(dtype))
        reservoir.set_param("bias", bias.astype(dtype))
        reservoir.set_param("internal_state", reservoir.zero_state())


def initialize_feedback(
    reservoir,
    feedback=None,
    Wfb_init=None,
    fb_scaling=None,
    fb_connectivity=None,
    seed=None,
):
    if reservoir.has_feedback:
        fb_dim = feedback.shape[1]
        reservoir.set_feedback_dim(fb_dim)

        if is_array(Wfb_init):
            Wfb = Wfb_init
            if not fb_dim == Wfb.shape[1]:
                raise ValueError(
                    "Dimension mismatch between Wfb and feedback "
                    f"vector in {reservoir.name}: Wfb is "
                    f"{Wfb.shape} "
                    f"and feedback is {(1, fb_dim)} "
                    f"({fb_dim} != {Wfb.shape[0]})"
                )

            if not Wfb.shape[0] == reservoir.output_dim:
                raise ValueError(
                    f"Dimension mismatch between Wfb and W in "
                    f"{reservoir.name}: Wfb is {Wfb.shape} and "
                    f"W is "
                    f"{reservoir.W.shape} ({Wfb.shape[1]} "
                    f"!= {reservoir.output_dim})"
                )

        elif callable(Wfb_init):
            Wfb = Wfb_init(
                reservoir.output_dim,
                fb_dim,
                input_scaling=fb_scaling,
                connectivity=fb_connectivity,
                seed=seed,
                dtype=reservoir.dtype,
            )
        else:
            raise ValueError(
                f"Data type {type(Wfb_init)} not understood "
                f"for matrix initializer 'Wfb_init' in "
                f"{reservoir.name}. Wfb should be an array "
                f"or a callable returning an array."
            )

        reservoir.set_param("Wfb", Wfb)



# Author: Nathan Trouvain at 16/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial

import numpy as np
from scipy import linalg

from ...mat_gen import zeros
from ...node import Node
from ...type import global_dtype
from .base import _initialize_readout, _prepare_inputs_for_learning, readout_forward


def _solve_ridge(XXT, YXT, ridge):
    """Solve Tikhonov regression."""
    return linalg.solve(XXT + ridge, YXT.T, assume_a="sym")


def _accumulate(readout, xxt, yxt):
    """Aggregate Xi.Xi^T and Yi.Xi^T matrices from a state sequence i."""
    XXT = readout.get_buffer("XXT")
    YXT = readout.get_buffer("YXT")
    XXT += xxt
    YXT += yxt


def partial_backward(readout: Node, X_batch, Y_batch=None, lock=None):
    """Pre-compute XXt and YXt before final fit."""
    X, Y = _prepare_inputs_for_learning(
        X_batch,
        Y_batch,
        bias=readout.input_bias,
        allow_reshape=True,
    )

    xxt = X.T.dot(X)
    yxt = Y.T.dot(X)

    if lock is not None:
        # This is not thread-safe using Numpy memmap as buffers
        # ok for parallelization then with a lock (see ESN object)
        with lock:
            _accumulate(readout, xxt, yxt)
    else:
        _accumulate(readout, xxt, yxt)


def backward(readout: Node, *args, **kwargs):
    ridge = readout.ridge
    XXT = readout.get_buffer("XXT")
    YXT = readout.get_buffer("YXT")

    input_dim = readout.input_dim
    if readout.input_bias:
        input_dim += 1

    ridgeid = ridge * np.eye(input_dim, dtype=global_dtype)

    Wout_raw = _solve_ridge(XXT, YXT, ridgeid)

    if readout.input_bias:
        Wout, bias = Wout_raw[1:, :], Wout_raw[0, :][np.newaxis, :]
        readout.set_param("Wout", Wout)
        readout.set_param("bias", bias)
    else:
        readout.set_param("Wout", Wout_raw)


def initialize(readout: Node, x=None, y=None, bias_init=None, Wout_init=None):
    _initialize_readout(
        readout, x, y, bias=readout.input_bias, init_func=Wout_init, bias_init=bias_init
    )


def initialize_buffers(readout):
    """create memmaped buffers for matrices X.X^T and Y.X^T pre-computed
    in parallel for ridge regression
    ! only memmap can be used ! Impossible to share Numpy arrays with
    different processes in r/w mode otherwise (with proper locking)
    """
    input_dim = readout.input_dim
    output_dim = readout.output_dim

    if readout.input_bias:
        input_dim += 1

    readout.create_buffer("XXT", (input_dim, input_dim))
    readout.create_buffer("YXT", (output_dim, input_dim))


class Ridge(Node):
    """A single layer of neurons learning with Tikhonov linear regression.

    Output weights of the layer are computed following:

    .. math::

        \\hat{\\mathbf{W}}_{out} = \\mathbf{YX}^\\top ~ (\\mathbf{XX}^\\top +
        \\lambda\\mathbf{Id})^{-1}

    Outputs :math:`\\mathbf{y}` of the node are the result of:

    .. math::

        \\mathbf{y} = \\mathbf{W}_{out}^\\top \\mathbf{x} + \\mathbf{b}

    where:
        - :math:`\\mathbf{X}` is the accumulation of all inputs during training;
        - :math:`\\mathbf{Y}` is the accumulation of all targets during training;
        - :math:`\\mathbf{b}` is the first row of :math:`\\hat{\\mathbf{W}}_{out}`;
        - :math:`\\mathbf{W}_{out}` is the rest of :math:`\\hat{\\mathbf{W}}_{out}`.

    If ``input_bias`` is True, then :math:`\\mathbf{b}` is non-zero, and a constant
    term is added to :math:`\\mathbf{X}` to compute it.

    :py:attr:`Ridge.params` **list**

    ================== =================================================================
    ``Wout``           Learned output weights (:math:`\\mathbf{W}_{out}`).
    ``bias``           Learned bias (:math:`\\mathbf{b}`).
    ================== =================================================================

    :py:attr:`Ridge.hypers` **list**

    ================== =================================================================
    ``ridge``          Regularization parameter (:math:`\\lambda`) (0.0 by default).
    ``input_bias``     If True, learn a bias term (True by default).
    ================== =================================================================


    Parameters
    ----------
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    ridge: float, default to 0.0
        L2 regularization parameter.
    Wout : callable or array-like of shape (units, targets), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    name : str, optional
        Node name.

    Example
    -------

    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.1, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.
    >>>
    >>> from reservoirpy.nodes import Ridge
    >>> ridge_regressor = Ridge(ridge=0.001)
    >>>
    >>> ridge_regressor.fit(x, y)
    >>> ridge_regressor.Wout, ridge_regressor.bias
    array([[ 9.992, -0.205,  6.989]]).T, array([[12.011]])
    """

    def __init__(
        self,
        output_dim=None,
        ridge=0.0,
        Wout=zeros,
        bias=zeros,
        input_bias=True,
        name=None,
    ):
        super(Ridge, self).__init__(
            params={"Wout": None, "bias": None},
            hypers={"ridge": ridge, "input_bias": input_bias},
            forward=readout_forward,
            partial_backward=partial_backward,
            backward=backward,
            output_dim=output_dim,
            initializer=partial(initialize, Wout_init=Wout, bias_init=bias),
            buffers_initializer=initialize_buffers,
            name=name,
        )



"""
=========================================================
Node operations - link, merge (:py:mod:`reservoirpy.ops`)
=========================================================

Operations on :py:class:`~.Node` and :py:class:`~.Model`.

.. currentmodule:: reservoirpy.ops

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   link - Link Nodes into a Model.
   link_feedback - Link Nodes through feedback connections.
   merge - Merge Models.
"""

# Author: Nathan Trouvain at 25/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from itertools import product
from typing import Iterable, Sequence, Union
from uuid import uuid4

from ._base import DistantFeedback, _Node
from .model import FrozenModel, Model
from .nodes.concat import Concat
from .utils.graphflow import find_parents_and_children

_MULTI_INPUTS_OPS = (Concat,)


def concat_multi_inputs(nodes, edges):
    parents, _ = find_parents_and_children(edges)

    concatenated = {}
    new_nodes = set()
    new_edges = set()
    for node in nodes:
        indegree = len(parents[node])
        if indegree > 1 and type(node) not in _MULTI_INPUTS_OPS:
            # add a Concat node
            concat = Concat()

            new_nodes |= {concat, node}
            new_edges |= set([(p, concat) for p in parents[node]] + [(concat, node)])
            # add concatenated nodes to the registry
            concatenated.update({p.name: concat for p in parents[node]})
        else:
            new_nodes |= {node}
            new_edges |= set([(p, node) for p in parents[node]])

    return list(new_nodes), list(new_edges)


def _check_all_nodes(*nodes):
    msg = "Impossible to link nodes: object {} is neither a Node nor a Model."
    for nn in nodes:
        if isinstance(nn, Iterable):
            for n in nn:
                if not isinstance(n, _Node):
                    raise TypeError(msg.format(n))
        else:
            if not isinstance(nn, _Node):
                raise TypeError(msg.format(nn))


def _link_1to1(node1: _Node, node2: _Node):
    """Connects two nodes or models. See `link` doc for more info."""
    # fetch all nodes in the two subgraphs, if they are models.
    all_nodes = []
    for node in (node1, node2):
        if isinstance(node, Model) and not isinstance(node, FrozenModel):
            all_nodes += node.nodes
        else:
            all_nodes += [node]

    # fetch all edges in the two subgraphs, if they are models.
    all_edges = []
    for node in (node1, node2):
        if isinstance(node, Model) and not isinstance(node, FrozenModel):
            all_edges += node.edges

    # create edges between output nodes of the
    # subgraph 1 and input nodes of the subgraph 2.
    senders = []
    if isinstance(node1, Model) and not isinstance(node, FrozenModel):
        senders += node1.output_nodes
    else:
        senders += [node1]

    receivers = []
    if isinstance(node2, Model) and not isinstance(node, FrozenModel):
        receivers += node2.input_nodes
    else:
        receivers += [node2]

    new_edges = list(product(senders, receivers))

    # maybe nodes are already initialized?
    # check if connected dimensions are ok
    for sender, receiver in new_edges:
        if (
            sender.is_initialized
            and receiver.is_initialized
            and sender.output_dim != receiver.input_dim
        ):
            raise ValueError(
                f"Dimension mismatch between connected nodes: "
                f"sender node {sender.name} has output dimension "
                f"{sender.output_dim} but receiver node "
                f"{receiver.name} has input dimension "
                f"{receiver.input_dim}."
            )

    # all outputs from subgraph 1 are connected to
    # all inputs from subgraph 2.
    all_edges += new_edges

    return all_nodes, all_edges


def link(
    node1: Union[_Node, Sequence[_Node]],
    node2: Union[_Node, Sequence[_Node]],
    name: str = None,
) -> Model:
    """Link two :py:class:`~.Node` instances to form a :py:class:`~.Model`
    instance. `node1` output will be used as input for `node2` in the
    created model. This is similar to a function composition operation:

    .. math::

        model(x) = (node1 \\circ node2)(x) = node2(node1(x))

    You can also perform this operation using the ``>>`` operator::

        model = node1 >> node2

    Or using this function::

        model = link(node1, node2)

    -`node1` and `node2` can also be :py:class:`~.Model` instances. In this
    case, the new :py:class:`~.Model` created will contain all nodes previously
    contained in all the models, and link all `node1` outputs to all `node2`
    inputs. This allows to chain the  ``>>`` operator::

        step1 = node0 >> node1  # this is a model
        step2 = step1 >> node2  # this is another

    -`node1` and `node2` can finally be lists or tuples of nodes. In this
    case, all `node1` outputs will be linked to a :py:class:`~.Concat` node to
    concatenate them, and the :py:class:`~.Concat` node will be linked to all
    `node2` inputs. You can still use the ``>>`` operator in this situation,
    except for many-to-many nodes connections::

        # many-to-one
        model = [node1, node2, ..., node] >> node_out
        # one-to-many
        model = node_in >> [node1, node2, ..., node]
        # ! many-to-many requires to use the `link` method explicitly!
        model = link([node1, node2, ..., node], [node1, node2, ..., node])

    Parameters
    ----------
        node1, node2 : _Node or list of _Node
            Nodes or lists of nodes to link.
        name: str, optional
            Name for the chaining Model.

    Returns
    -------
        Model
            A :py:class:`~.Model` instance chaining the nodes.

    Raises
    ------
        TypeError
            Dimension mismatch between connected nodes: `node1` output
            dimension if different from `node2` input dimension.
            Reinitialize the nodes or create new ones.

    Notes
    -----

        Be careful to how you link the different nodes: `reservoirpy` does not
        allow to have circular dependencies between them::

            model = node1 >> node2  # fine
            model = node1 >> node2 >> node1  # raises! data would flow in
                                             # circles forever...
    """

    _check_all_nodes(node1, node2)

    frozens = []
    if isinstance(node1, Sequence):
        frozens += [n.name for n in node1 if isinstance(n, FrozenModel)]
    else:
        if isinstance(node1, FrozenModel):
            frozens.append(node1)
    if isinstance(node2, Sequence):
        frozens += [n.name for n in node2 if isinstance(n, FrozenModel)]
    else:
        if isinstance(node2, FrozenModel):
            frozens.append(node2)

    if len(frozens) > 0:
        raise TypeError(
            "Impossible to link FrozenModel to other Nodes or "
            f"Model. FrozenModel found: {frozens}."
        )

    nodes = set()
    edges = set()
    if not isinstance(node1, Sequence):
        node1 = [node1]
    if not isinstance(node2, Sequence):
        node2 = [node2]

    for left in node1:
        for right in node2:
            new_nodes, new_edges = _link_1to1(left, right)
            nodes |= set(new_nodes)
            edges |= set(new_edges)

    return Model(nodes=list(nodes), edges=list(edges), name=name)


def link_feedback(
    node: _Node,
    feedback: Union[_Node, Sequence[_Node]],
    inplace: bool = False,
    name: str = None,
) -> _Node:
    """Create a feedback connection between the `feedback` node and `node`.
    Feedbacks nodes will be called at runtime using data from the previous
    call.

    This is not an in-place operation by default. This function will copy `node`
    and then sets the copy `_feedback` attribute as a reference to `feedback`
    node. If `inplace` is set to `True`, then `node` is not copied and the
    feedback is directly connected to `node`. If `feedback` is a list of nodes
    or models, then all nodes in the list are first connected to a
    :py:class:`~.Concat` node to create a model gathering all data from all nodes
    in a single feedback vector.

     You can also perform this operation using the ``<<`` operator::

        node1 = node1 << node2
        # with feedback from a Model
        node1 = node1 << (fbnode1 >> fbnode2)
        # with feedback from a list of nodes or models
        node1 = node1 << [fbnode1, fbnode2, ...]

    Which means that a feedback connection is now created between `node1` and
    `node2`. In other words, the forward function of `node1` depends on the
    previous output of `node2`:

    .. math::
        \\mathrm{node1}(x_t) = \\mathrm{node1}(x_t, \\mathrm{node2}(x_{t - 1}))

    You can also use this function to define feedback::

        node1 = link_feedback(node1, node2)
        # without copy (node1 is the same object throughout)
        node1 = link_feedback(node1, node2, inplace=True, name="n1_copy")

    Parameters
    ----------
    node : Node
        Node receiving feedback.
    feedback : _Node
        Node or Model sending feedback
    inplace : bool, defaults to False
        If `True`, then the function returns a copy of `node`.
    name : str, optional
        Name of the copy of `node` if `inplace` is `True`.

    Returns
    -------
        Node
            A node instance taking feedback from `feedback`.

    Raises
    ------
        TypeError
            - If `node` is a :py:class:`~.Model`.
            Models can not receive feedback.

            - If any of the feedback nodes are not :py:class:`~._Node`
            instances.
    """

    if isinstance(node, Model):
        raise TypeError(f"{node} is not a Node. Models can't receive feedback.")

    msg = (
        "Impossible to receive feedback from {}: "
        "it is not a Node or a Model instance."
    )

    if isinstance(feedback, Sequence):
        for fb in feedback:
            if not isinstance(fb, _Node):
                raise TypeError(msg.format(fb))

        all_fb = link(feedback, Concat())

    elif isinstance(feedback, _Node):
        all_fb = feedback

    else:
        raise TypeError(msg.format(feedback))

    if inplace:
        node._feedback = DistantFeedback(all_fb, node)
        return node
    else:
        # first copy the node, then give it feedback
        # original node is not connected to any feedback then
        new_node = node.copy(name=name)
        new_node._feedback = DistantFeedback(all_fb, new_node)
        return new_node


def merge(
    model: _Node, *models: _Node, inplace: bool = False, name: str = None
) -> Model:
    """Merge different :py:class:`~.Model` or :py:class:`~.Node`
    instances into a single :py:class:`~.Model` instance.

    :py:class:`~.Node` instances contained in the models to merge will be
    gathered in a single model, along with all previously defined connections
    between them, if they exists.

    You can also perform this operation using the ``&`` operator::

        model = (node1 >> node2) & (node1 >> node3))

    This is equivalent to::

        model = merge((node1 >> node2), (node1 >> node3))

    The in-place operator can also be used::

        model &= other_model

    Which is equivalent to::

        model.update_graph(other_model.nodes, other_model.edges)

    Parameters
    ----------
    model: Model or Node
        First node or model to merge. The `inplace` parameter takes this
        instance as reference.
    *models : Model or Node
        All models to merge.
    inplace: bool, default to False
        If `True`, then will update Model `model` in-place. If `model` is not
        a Model instance, this parameter will causes the function to raise
        a `ValueError`.
    name: str, optional
        Name of the resulting Model.

    Returns
    -------
    Model
        A new :py:class:`~.Model` instance.

    Raises
    ------
    ValueError
        If `inplace` is `True` but `model` is not a Model instance, then the
        operation is impossible. In-place merging can only take place on a
        Model instance.
    """
    msg = "Impossible to merge models: object {} is not a Model instance."

    if isinstance(model, _Node):
        all_nodes = set()
        all_edges = set()
        for m in models:
            # fuse models nodes and edges (right side argument)
            if isinstance(m, Model) and not isinstance(m, FrozenModel):
                all_nodes |= set(m.nodes)
                all_edges |= set(m.edges)
            elif isinstance(m, _Node):
                all_nodes |= {m}

        if inplace:
            if not isinstance(model, Model) or isinstance(model, FrozenModel):
                raise ValueError(
                    f"Impossible to merge models in-place: "
                    f"{model} is not a Model instance."
                )
            return model.update_graph(all_nodes, all_edges)

        else:
            # add left side model nodes
            if isinstance(model, Model) and not isinstance(model, FrozenModel):
                all_nodes |= set(model.nodes)
                all_edges |= set(model.edges)
            else:
                all_nodes |= {model}

            return Model(nodes=list(all_nodes), edges=list(all_edges), name=name)

    else:
        raise TypeError(msg.format(type(model)))



# Author: Nathan Trouvain at 15/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import warnings

import numpy as np

try:
    import sklearn
except ImportError:
    sklearn = None

from ..node import Node


def forward_predict(estimator, x):
    return np.atleast_2d(estimator.estimator.predict(x))


def forward_transform(estimator, x):
    return np.atleast_2d(estimator.estimator.transform(x))


def backward(estimator, X, Y=None):
    estimator.hypers["estimator"] = estimator.estimator.fit(X, Y)


def initialize(estimator, x, y=None):
    ...


class from_sklearn(Node):
    def __init__(self, estimator):
        warnings.warn(
            "This experimental node is deprecated and should not be used. "
            "Please consider using the reservoirpy.nodes.ScikitLearnNode instead.",
            DeprecationWarning,
        )

        if hasattr(estimator, "predict"):
            forward = forward_predict
        elif hasattr(estimator, "transform"):
            forward = forward_transform
        else:
            raise TypeError(
                f"Estimator {estimator} has no 'predict' or 'transform' attribute."
            )

        super(from_sklearn, self).__init__(
            hypers={"estimator": estimator}, forward=forward, backward=backward
        )



# Author: Nathan Trouvain at 16/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import warnings
from functools import partial
from numbers import Number
from typing import Callable, Iterable, Union

from ...mat_gen import zeros
from ...node import Node
from ...type import Weights
from .base import readout_forward
from .lms import initialize as initialize_lms
from .lms import train as lms_like_train
from .rls import initialize as initialize_rls
from .rls import train as rls_like_train

RULES = ("lms", "rls")


class FORCE(Node):
    """Single layer of neurons learning connections through online learning rules.

    Warning
    -------

    This class is deprecated since v0.3.4 and will be removed in future versions.
    Please use :py:class:`~reservoirpy.LMS` or :py:class:`~reservoirpy.RLS` instead.


    The learning rules involved are similar to Recursive Least Squares (``rls`` rule)
    as described in [1]_ or Least Mean Squares (``lms`` rule, similar to Hebbian
    learning) as described in [2]_.

    "FORCE" name refers to the training paradigm described in [1]_.

    :py:attr:`FORCE.params` **list**

    ================== =================================================================
    ``Wout``           Learned output weights (:math:`\\mathbf{W}_{out}`).
    ``bias``           Learned bias (:math:`\\mathbf{b}`).
    ``P``              Matrix :math:`\\mathbf{P}` of RLS rule (optional).
    ================== =================================================================

    :py:attr:`FORCE.hypers` **list**

    ================== =================================================================
    ``alpha``          Learning rate (:math:`\\alpha`) (:math:`1\\cdot 10^{-6}` by
                       default).
    ``input_bias``     If True, learn a bias term (True by default).
    ``rule``           One of RLS or LMS rule ("rls" by default).
     ``forgetting``    Forgetting factor, only used with RLS (:math:`\\lambda`)
                       (:math:`1` by default).
    ================== =================================================================

    Parameters
    ----------
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    alpha : float or Python generator or iterable, default to 1e-6
        Learning rate. If an iterable or a generator is provided and the learning
        rule is "lms", then the learning rate can be changed at each timestep of
        training. A new learning rate will be drawn from the iterable or generator
        at each timestep.
    rule : {"rls", "lms"}, default to "rls"
        Learning rule applied for online training.
    Wout : callable or array-like of shape (units, targets), default to
        :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to
        :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    forgetting : float, default to 1.0
        Forgetting factor for the RLS rule. Ignored if rule is "lms".
    name : str, optional
        Node name.

    References
    ----------

    .. [1] Sussillo, D., & Abbott, L. F. (2009). Generating Coherent Patterns of
           Activity from Chaotic Neural Networks. Neuron, 63(4), 544–557.
           https://doi.org/10.1016/j.neuron.2009.07.018

    .. [2] Hoerzer, G. M., Legenstein, R., & Maass, W. (2014). Emergence of Complex
           Computational Structures From Chaotic Neural Networks Through
           Reward-Modulated Hebbian Learning. Cerebral Cortex, 24(3), 677–690.
           https://doi.org/10.1093/cercor/bhs348
    """

    def __init__(
        self,
        output_dim: int = None,
        alpha: Union[float, Iterable[float]] = 1e-6,
        rule: str = "rls",
        Wout: Union[Weights, Callable] = zeros,
        bias: Union[Weights, Callable] = zeros,
        input_bias: bool = True,
        forgetting: float = 1.0,
        name: str = None,
    ):

        warnings.warn(
            "'FORCE' is deprecated since v0.3.4 and will be removed "
            "in "
            "future versions. Consider using 'RLS' or 'LMS'.",
            DeprecationWarning,
        )

        params = {"Wout": None, "bias": None}

        if rule not in RULES:
            raise ValueError(
                f"Unknown rule for FORCE learning. "
                f"Available rules are {self._rules}."
            )
        else:
            if rule == "lms":
                train = lms_like_train
                initialize = initialize_lms
            else:
                train = rls_like_train
                initialize = initialize_rls
                params["P"] = None

        if isinstance(alpha, Number):

            def _alpha_gen():
                while True:
                    yield alpha

            alpha_gen = _alpha_gen()
        elif isinstance(alpha, Iterable):
            alpha_gen = alpha

        hypers = {
            "alpha": alpha,
            "_alpha_gen": alpha_gen,
            "input_bias": input_bias,
            "rule": rule,
            "forgetting": forgetting,
        }

        super(FORCE, self).__init__(
            params=params,
            hypers=hypers,
            forward=readout_forward,
            train=train,
            initializer=partial(
                initialize, init_func=Wout, bias_init=bias, bias=input_bias
            ),
            output_dim=output_dim,
            name=name,
        )
